from relationalai.early_access.dsl.core.types import AbstractValueType
from relationalai.early_access.dsl.core.types.variables import BasicTypeVariable
import numbers

class UnconstrainedValueType(AbstractValueType):

    def __init__(self, nm):
        AbstractValueType.__init__(self, nm)

    # Returns True if the Python constant `val` is an instance of this ValueType
    def contains(self, val) -> bool:
        if isinstance(val, numbers.Number):
            return self.basic_numeric()
        elif isinstance(val, str):
            return not self.basic_numeric()
        else:
            return False

    def pprint(self): return self.display()

    def ranges_over(self): return BasicTypeVariable(self)

    def relational(self): return False

    def root_unconstrained_type(self): return self

    def unconstrained(self): return True

class UnconstrainedNumericType(UnconstrainedValueType):

    def __init__(self, nm):
        UnconstrainedValueType.__init__(self, nm)

    def boolean(self): return self.display() == 'Boolean'

    # For verification purposes, we treat floats ad decimals. While not generally
    # safe, this is likely fine for constraints we will see in enterprise contexts.
    #
    def decimal(self): return self.display() == 'Decimal' or self.display() == 'Float'

    def integer(self): return self.display() == 'Integer' or self.display() == 'BigInteger'

    def basic_numeric(self) -> bool: return True

class UnsupportedUnconstrainedValueType(UnconstrainedValueType):
    def __init__(self, nm):
        UnconstrainedValueType.__init__(self, nm)

    def root_unconstrained_type(self):
        raise Exception(f"Unsupported basic type {self.display()} cannot be the base type of any value type")



from __future__ import annotations
from typing import Union, Literal
import datetime as dt

from .std import _DateTime, _Date, _Integer, _String, _make_expr
from .. import std
from .. import builder as b

# TODO support DateTime below as well, but this needs e.g. Rel `datetime_year`

def year(date: _Date) -> b.Expression:
    return _make_expr("date_year", date, b.Int64.ref("res"))

def month(date: _Date) -> b.Expression:
    return _make_expr("date_month", date, b.Int64.ref("res"))

def day(date: _Date) -> b.Expression:
    return _make_expr("date_day", date, b.Int64.ref("res"))

def dates_period_days(start: _Date, end: _Date) -> b.Expression:
    return _make_expr("dates_period_days", start, end, b.Int64.ref("res"))

def datetimes_period_milliseconds(start: _DateTime, end: _DateTime) -> b.Expression:
    return _make_expr("datetimes_period_milliseconds", start, end, b.Int64.ref("res"))
#--------------------------------------------------
# Periods
#--------------------------------------------------
def milliseconds(period: _Integer) -> b.Expression:
    return _make_expr("millisecond", period, b.Int64.ref("res"))

def seconds(period: _Integer) -> b.Expression:
    return _make_expr("second", period, b.Int64.ref("res"))

def minutes(period: _Integer) -> b.Expression:
    return _make_expr("minute", period, b.Int64.ref("res"))

def hours(period: _Integer) -> b.Expression:
    return _make_expr("hour", period, b.Int64.ref("res"))

def days(period: _Integer) -> b.Expression:
    return _make_expr("day", period, b.Int64.ref("res"))

def weeks(period: _Integer) -> b.Expression:
    return _make_expr("week", period, b.Int64.ref("res"))

def months(period: _Integer) -> b.Expression:
    return _make_expr("month", period, b.Int64.ref("res"))

def years(period: _Integer) -> b.Expression:
    return _make_expr("year", period, b.Int64.ref("res"))

def date_to_datetime(date: _Date, hour: int = 0, minute: int = 0, second: int = 0, millisecond: int = 0, tz: str = "UTC") -> b.Expression:
    _year = year(date)
    _month = month(date)
    _day = day(date)
    return _make_expr("construct_datetime_ms_tz", _year, _month, _day, hour, minute, second, millisecond, tz, b.DateTime.ref("res"))

def datetime_to_date(datetime: _DateTime, tz: dt.tzinfo | str | None = None) -> b.Expression:
    if tz is None:
        if isinstance(datetime, dt.datetime):
            tz = datetime.tzname() or "UTC"
        else:
            tz = "UTC"

    return _make_expr("construct_date_from_datetime", datetime, tz, b.Date.ref("res"))

#--------------------------------------------------
# String Formatting
#--------------------------------------------------

def date_format(date: _Date, format: _String) -> b.Expression:
    return _make_expr("date_format", date, format, b.String.ref("res"))

def datetime_format(date: _DateTime, format: _String, tz: _String = "UTC") -> b.Expression:
    return _make_expr("datetime_format", date, format, tz, b.String.ref("res"))

#--------------------------------------------------
# Arithmetic
#--------------------------------------------------
def date_add(date: _Date, period: b.Producer) -> b.Expression:
    return _make_expr("date_add", date, period, b.Date.ref("res"))

def date_subtract(date: _Date, period: b.Producer) -> b.Expression:
    return _make_expr("date_subtract", date, period, b.Date.ref("res"))

def datetime_add(date: _DateTime, period: b.Producer) -> b.Expression:
    return _make_expr("datetime_add", date, period, b.DateTime.ref("res"))

def datetime_subtract(date: _DateTime, period: b.Producer) -> b.Expression:
    return _make_expr("datetime_subtract", date, period, b.DateTime.ref("res"))


Frequency = Union[
    Literal["ms"],
    Literal["s"],
    Literal["m"],
    Literal["H"],
    Literal["D"],
    Literal["W"],
    Literal["M"],
    Literal["Y"],
]

_periods = {
    "ms": milliseconds,
    "s": seconds,
    "m": minutes,
    "H": hours,
    "D": days,
    "W": weeks,
    "M": months,
    "Y": years,
}

# Note on date_ranges and datetime_range: The way the computation works is that it first overapproximates the number of periods.
# For example date_range(2025-02-01, 2025-03-01, freq='M') and date_range(2025-02-01, 2025-03-31, freq='M') will compute
# range_end to be ceil(28*1/(365/12))=1 and ceil(58*1/(365/12))=2. Then, the computation fetches range_end+1 items into _date, which
# is the right number in the first case but one too many in the second case. That's why a filter end >= _date (or variant of) is
# applied, to remove any extra item. The result is two items in both cases.

def date_range(start: _Date | None = None, end: _Date | None = None, periods: int = 1, freq: Frequency = "D") -> b.Expression:
    if start is None and end is None:
        raise ValueError("Invalid start/end date for date_range. Must provide at least start date or end date")
    _days = {
        "D": 1,
        "W": 1/7,
        "M": 1/(365/12),
        "Y": 1/365,
    }
    if freq not in _days.keys():
        raise ValueError(f"Frequency '{freq}' is not allowed for date_range. List of allowed frequencies: {list(_days.keys())}")
    date_func = date_add
    if start is None:
        start = end
        end = None
        date_func = date_subtract
    assert start is not None
    if end is not None:
        num_days = std.dates.dates_period_days(start, end)
        if freq in ["W", "M", "Y"]:
            range_end = std.cast(b.Int64, std.math.ceil(num_days * _days[freq]))
        else:
            range_end = num_days
        # date_range is inclusive. add 1 since std.range is exclusive
        ix = std.range(0, range_end + 1, 1)
    else:
        ix = std.range(0, periods, 1)
    _date = date_func(start, _periods[freq](ix))
    if isinstance(end, dt.date) :
        return b.Date(end) >= _date
    elif end is not None:
        return end >= _date
    return _date

def datetime_range(start: _DateTime | None = None, end: _DateTime | None = None, periods: int = 1, freq: Frequency = "D") -> b.Expression:
    if start is None and end is None:
        raise ValueError("Invalid start/end datetime for datetime_range. Must provide at least start datetime or end datetime")
    _milliseconds = {
        "ms": 1,
        "s": 1 / 1_000,
        "m": 1 / 60_000,
        "H": 1 / 3_600_000,
        "D": 1 / 86_400_000,
        "W": 1 / (86_400_000 * 7),
        "M": 1 / (86_400_000 * (365 / 12)),
        "Y": 1 / (86_400_000 * 365),
    }
    date_func = datetime_add
    if start is None:
        start = end
        end = None
        date_func = datetime_subtract
    assert start is not None
    if end is not None:
        num_ms = datetimes_period_milliseconds(start, end)
        if freq == "ms":
            _end = num_ms
        else:
            _end = std.cast(b.Int64, std.math.ceil(num_ms * _milliseconds[freq]))
        # datetime_range is inclusive. add 1 since std.range is exclusive
        ix = std.range(0, _end + 1, 1)
    else:
        ix = std.range(0, periods, 1)
    _date = date_func(start, _periods[freq](ix))
    if isinstance(end, dt.datetime) :
        return b.DateTime(end) >= _date
    elif end is not None:
        return end >= _date
    return _date



from __future__ import annotations

from dataclasses import dataclass
from relationalai.early_access.metamodel import ir, compiler as c, factory as f, visitor


class DoubleNegation(c.Pass):
    """
        Pass to drop double negations.

        Examples:
        not_(not_(Person.age > 18, Person.id == 2)) -> (Person.age > 18, Person.id == 2)
        not_(not_(not_(Person.age > 18, Person.id == 2))) -> not_(Person.age > 18, Person.id == 2)
    """

    def rewrite(self, model: ir.Model, options: dict = {}) -> ir.Model:
        return DropDoubleNegationsVisitor().walk(model)


@dataclass
class DropDoubleNegationsVisitor(visitor.Rewriter):
    """
    Visitor drops double negations when they are followed one by one in IR Model.
    The rewrite occurs only when `ir.Not`'s task is an `ir.Logical` with single `ir.Not` in the body.

    Example:
        Not
            Logical
                Not
                    Logical
                        age(person, age)
                        age > 18

        Logical
            age(person, age)
            age > 18
    """

    def handle_not(self, node: ir.Not, parent: ir.Node):
        if isinstance(node.task, ir.Logical) and len(node.task.body) == 1 and isinstance(node.task.body[0], ir.Not):
            new_task = self.walk(node.task.body[0].task, node.task.body[0])
            # dropping redundant Logical if it has only 1 element in his body
            if isinstance(new_task, ir.Logical) and len(new_task.body) == 1:
                return new_task.body[0]
            else:
                return new_task
        else:
            new_task = self.walk(node.task, node)
            return node if node.task is new_task else f.not_(new_task)



from typing import Iterable, Dict, Optional, List, cast, TYPE_CHECKING
import json
import uuid

from relationalai import debugging
from relationalai.clients.cache_store import GraphIndexCache
from relationalai.clients.util import get_pyrel_version, poll_with_specified_overhead
from relationalai.errors import (
    ERPNotRunningError,
    EngineAutoCreateFailed,
    SnowflakeChangeTrackingNotEnabledException,
    SnowflakeTableObjectsException,
    SnowflakeTableObject,
)
from relationalai.tools.cli_controls import DebuggingSpan, Spinner
from relationalai.tools.constants import WAIT_FOR_STREAM_SYNC, Generation

if TYPE_CHECKING:
    from relationalai.clients.snowflake import Resources
    from relationalai.clients.snowflake import DirectAccessResources


class UseIndexPoller:
    """
    Encapsulates the polling logic for `use_index` streams.
    """
    def __init__(
        self,
        resource: "Resources",
        app_name: str,
        sources: Iterable[str],
        model: str,
        engine_name: str,
        engine_size: Optional[str],
        program_span_id: Optional[str],
        headers: Optional[Dict],
        generation: Optional[Generation] = None,
    ):
        self.res = resource
        self.app_name = app_name
        self.sources = list(sources)
        self.model = model
        self.engine_name = engine_name
        self.engine_size = engine_size or self.res.config.get_default_engine_size()
        self.program_span_id = program_span_id
        self.headers = headers or {}
        self.counter = 1
        self.check_ready_count = 0
        self.tables_with_not_enabled_change_tracking: List = []
        self.table_objects_with_other_errors: List = []
        self.engine_errors: List = []
        # Flag to only ensure the engine is created asynchronously the initial call
        self.init_engine_async = True
        # Initially, we assume that cdc is not checked,
        # then on subsequent calls, if we get if cdc is enabled, if it is not, we will check it
        # on every 5th iteration we reset the cdc status, so it will be checked again
        self.should_check_cdc = True

        self.wait_for_stream_sync = self.res.config.get(
            "wait_for_stream_sync", WAIT_FOR_STREAM_SYNC
        )
        current_user = self.res.get_sf_session().get_current_user()
        assert current_user is not None, "current_user must be set"
        data_freshness = self.res.config.get_data_freshness_mins()
        self.cache = GraphIndexCache(current_user, model, data_freshness, self.sources)
        self.sources = self.cache.choose_sources()
        # execution_id is allowed to group use_index call, which belongs to the same loop iteration
        self.execution_id = str(uuid.uuid4())

        self.pyrel_version = get_pyrel_version(generation)

        self.source_info = self.res._check_source_updates(self.sources)

    def poll(self) -> None:
        """
        Standard stream-based polling for use_index.
        """
        with Spinner(
            "Initializing data index... ",
            "Setup complete",
            leading_newline=True,
            trailing_newline=True,
        ) as spinner:
            spinner.update_messages({ "message": "Validating data sources..." })
            self._maybe_delete_stale(spinner)
            spinner.update_messages({ "message": "Indexing your data in the background. (This may take a while)..." })
            self._poll_loop(spinner)
            self._post_check(spinner)

    def _maybe_delete_stale(self, spinner) -> None:
        with debugging.span("check_sources"):
            # Source tables that have been altered/changed since the last stream creation
            stale_sources = [
                source
                for source, info in self.source_info.items()
                if info["state"] == "STALE"
            ]
        if stale_sources:
            with DebuggingSpan("validate_sources"):
                try:
                    # Delete all stale streams, so use_index could recreate them again
                    from relationalai.clients.snowflake import PYREL_ROOT_DB
                    query = f"CALL {self.app_name}.api.delete_data_streams({stale_sources}, '{PYREL_ROOT_DB}');"
                    delete_response = self.res._exec(query)
                    delete_json_str = delete_response[0]["DELETE_DATA_STREAMS"].lower()
                    delete_data = json.loads(delete_json_str)
                    diff = len(stale_sources) - delete_data.get("deleted", 0)
                    if diff > 0:
                        errors = delete_data.get("errors", None)
                        if errors:
                            raise Exception(f"Error(s) deleting streams with modified sources: {errors}")
                except Exception as e:
                    # The delete_data_streams procedure will raise an exception if the streams do not exist
                    if "data streams do not exist" in str(e).lower():
                        pass
                    else:
                        raise e from None

    def _poll_loop(self, spinner) -> None:
        source_references = self.res._get_source_references(self.source_info)
        sources_object_references_str = ", ".join(source_references)

        def check_ready(spinner) -> bool:
            self.check_ready_count += 1

            # To limit the performance overhead, we only check if ERP is running every 5 iterations
            if self.check_ready_count % 5 == 0:
                with debugging.span("check_erp_status"):
                    if not self.res.is_erp_running(self.app_name):
                        raise ERPNotRunningError

            use_index_id = f"{self.model}_{self.execution_id}"

            params = json.dumps({
                "model": self.model,
                "engine": self.engine_name,
                "default_engine_size": self.engine_size, # engine_size
                "user_agent": self.pyrel_version,
                "use_index_id": use_index_id,
                "pyrel_program_id": self.program_span_id,
                "wait_for_stream_sync": self.wait_for_stream_sync,
                "should_check_cdc": self.should_check_cdc,
                "init_engine_async": self.init_engine_async,
            })
            sql_string = f"CALL {self.app_name}.api.use_index([{sources_object_references_str}], PARSE_JSON(?), {self.headers});"

            with debugging.span("wait", counter=self.counter, use_index_id=use_index_id) as span:
                results = self.res._exec(sql_string, [params])

                # Extract the JSON string from the `USE_INDEX` field
                use_index_json_str = results[0]["USE_INDEX"]

                # Parse the JSON string into a Python dictionary
                use_index_data = json.loads(use_index_json_str)
                span.update(use_index_data)

                all_data = use_index_data.get("data", [])
                ready = use_index_data.get("ready", False)
                engines = use_index_data.get("engines", [])
                errors = use_index_data.get("errors", [])
                cdc_enabled = use_index_data.get("cdcEnabled", False)
                if self.check_ready_count % 5 == 0 or not cdc_enabled:
                    self.should_check_cdc = True
                else:
                    self.should_check_cdc = False

                break_loop = False

                if ready:
                    self.cache.record_update(self.source_info)
                    break_loop = True
                    spinner.update_messages({ "finished_message": "Setup complete" })

                if not ready and all_data:
                    for data in all_data:
                        status = data.get("status", "").lower()
                        if data.get("errors", []):
                            for error in data.get("errors", []):
                                error_msg = f"{error.get('error')}, source: {error.get('source')}"
                                self.table_objects_with_other_errors.append(
                                    SnowflakeTableObject(error_msg, data.get("fq_object_name"))
                                )
                            break_loop = True

                        if status == "created":
                            spinner.update_messages({ "message": f"Indexing {data.get('fq_object_name')}..." })
                        if (status != "synced" or data.get("pending_batches_count") > 0):
                            batches_count = data.get("pending_batches_count")
                            if batches_count:
                                spinner.update_messages({ "message": f"Indexing {data.get('fq_object_name')}, remaining batches: {data.get('pending_batches_count')}..." })
                            else:
                                spinner.update_messages({ "message": f"Indexing {data.get('fq_object_name')}..." })
                    self.counter += 1

                if not ready and engines:
                    for engine in engines:
                        name = engine.get("name", "")
                        state = engine.get("state", "").lower()
                        status = engine.get("status", "").lower()
                        if (state == "pending" or status == "pending"):
                            writer = engine.get("writer", False)
                            message = f"Waiting for {writer and 'writer engine' or 'engine'} {name}..."
                            spinner.update_messages({ "message": message })
                    self.counter += 1

                if not ready and errors:
                    for error in errors:
                        if error.get("type") == "data":
                            message = error.get("message", "").lower()
                            if ("change_tracking" in message or "change tracking" in message):
                                err_source = error.get("source")
                                err_source_type = self.source_info.get(err_source, {}).get("type")
                                self.tables_with_not_enabled_change_tracking.append((err_source, err_source_type))
                            else:
                                self.table_objects_with_other_errors.append(
                                    SnowflakeTableObject(error.get("message"), error.get("source"))
                                )
                        elif error.get("type") == "engine":
                            self.engine_errors.append(error)
                        else:
                            # Other types of errors, e.g. "validation"
                            self.table_objects_with_other_errors.append(
                                SnowflakeTableObject(error.get("message"), error.get("source"))
                            )
                    break_loop = True
                return break_loop

        poll_with_specified_overhead(lambda: check_ready(spinner), overhead_rate=0.1, max_delay=1)

    def _post_check(self, spinner) -> None:
            num_tables_altered = 0

            enabled_tables = []
            if (
                self.tables_with_not_enabled_change_tracking
                and self.res.config.get("ensure_change_tracking", False)
            ):
                for table in self.tables_with_not_enabled_change_tracking:
                    try:
                        fqn, kind = table
                        self.res._exec(f"ALTER {kind} {fqn} SET CHANGE_TRACKING = TRUE;")
                        enabled_tables.append(table)
                        num_tables_altered += 1
                    except Exception:
                        pass
                # Remove the tables that were successfully enabled from the list of not enabled tables
                # so that we don't raise an exception for them later
                self.tables_with_not_enabled_change_tracking = [
                    t for t in self.tables_with_not_enabled_change_tracking if t not in enabled_tables
                ]

            if self.tables_with_not_enabled_change_tracking:
                spinner.update_messages({ "message": "Errors found. See below for details." })
                raise SnowflakeChangeTrackingNotEnabledException(
                    self.tables_with_not_enabled_change_tracking
                )

            if self.table_objects_with_other_errors:
                spinner.update_messages({ "message": "Errors found. See below for details." })
                raise SnowflakeTableObjectsException(self.table_objects_with_other_errors)
            if self.engine_errors:
                spinner.update_messages({ "message": "Errors found. See below for details." })
                # if there is an engine error, probably auto create engine failed
                raise EngineAutoCreateFailed(self.engine_name)

            if num_tables_altered > 0:
                s = "s" if num_tables_altered > 1 else ""
                spinner.update_messages({ "message": f"Enabled change tracking on {num_tables_altered} table{s}..." })
                self._poll_loop(spinner)

class DirectUseIndexPoller(UseIndexPoller):
    """
    Extends UseIndexPoller to handle direct-access prepare_index when no sources.
    """
    def __init__(
        self,
        resource: "DirectAccessResources",
        app_name: str,
        sources: Iterable[str],
        model: str,
        engine_name: str,
        engine_size: Optional[str],
        program_span_id: Optional[str],
        headers: Optional[Dict],
        generation: Optional[Generation] = None,
    ):
        super().__init__(resource, app_name, sources, model, engine_name, engine_size, program_span_id, headers, generation)
        from relationalai.clients.snowflake import DirectAccessResources
        self.res: DirectAccessResources = cast(DirectAccessResources, self.res)

    def poll(self) -> None:
        if not self.sources:
            from relationalai.errors import RAIException
            collected_errors: List[Dict] = []
            attempt = 1

            def check_direct(spinner) -> bool:
                nonlocal attempt
                with debugging.span("wait", counter=self.counter) as span:
                    span.update({"attempt": attempt, "engine_name": self.engine_name, "model": self.model})
                    # we are skipping pulling relations here, as direct access only handle non-sources cases
                    # and we don't need to pull relations for that, therefore, we pass empty list for rai_relations
                    # and set skip_pull_relations to True
                    resp = self.res._prepare_index(
                        model=self.model,
                        engine_name=self.engine_name,
                        engine_size=self.engine_size,
                        rai_relations=[],
                        pyrel_program_id=self.program_span_id,
                        skip_pull_relations=True,
                        headers=self.headers,
                    )
                    span.update(resp)
                    caller_engine = resp.get("caller_engine", {})
                    ce_status = caller_engine.get("status", "").lower()
                    errors = resp.get("errors", [])

                    ready = resp.get("ready", False)

                    if ready:
                        spinner.update_messages({"finished_message": "Setup complete"})
                        return True
                    else:
                        if ce_status == "pending":
                            message = f"Waiting for engine '{caller_engine.get('name', self.engine_name)}' to be ready..."
                            spinner.update_messages({"message": message})
                        else:
                            for err in errors:
                                collected_errors.append(err)

                    attempt += 1
                    return False

            with Spinner(
                "Preparing your data...",
                "Setup complete",
                leading_newline=True,
                trailing_newline=True,
            ) as spinner:
                with debugging.span("poll_direct"):
                    poll_with_specified_overhead(lambda: check_direct(spinner), overhead_rate=0.1, max_delay=1)

            if collected_errors:
                spinner.update_messages({"message": "Errors found. See below for details."})
                msg = "; ".join(e.get("message", "") for e in collected_errors)
                raise RAIException(msg)
        else:
            super().poll()



import json
import re
import warnings

import xmltodict
from collections import defaultdict
from relationalai.early_access.dsl.adapters.orm.model import ORMEntityType, ORMValueType, ORMRole, ORMSubtypeFact, \
    ORMUniquenessConstraint, ORMExclusionConstraint, ORMMandatoryConstraint, ORMReadingRole, ORMReading, \
    ExclusiveInclusiveSubtypeFact, InclusiveSubtypeFact, ExclusiveSubtypeFact, SubtypeFact, ORMInclusionConstraint, \
    ORMRoleValueConstraint, ORMValueComparisonConstraint, ORMValueComparisonOperator, ORMRingConstraint, ORMRingType, \
    ORMRoleSubsetConstraint, ORMInclusiveRoleConstraint, ORMExclusiveRoleConstraint, ORMEqualityConstraint, \
    ORMFrequencyConstraint, ORMCardinalityConstraint, ORMRange, ORMRoleCardinalityConstraint, \
    ORMValueTypeValueConstraint
from relationalai.early_access.metamodel.util import ordered_set
from relationalai.util.graph import topological_sort


class ORMParser:

    def __init__(self, orm_file_path):
        with open(orm_file_path) as orm_file:
            data = xmltodict.parse(orm_file.read())

        self._ontology = json.loads(json.dumps(data))

        self._role_to_player = dict()
        self._cardinality_constraints = dict()
        self._role_cardinality_constraints = dict()
        self._value_type_value_constraints = dict()

        self._model_name = self._parse_model_name()
        self._entity_types = self._parse_entity_types()
        self._value_types = self._parse_value_types()
        self._object_types = {**self._value_types, **self._entity_types}
        self._roles, self._ignored_roles = self._parse_roles()
        self._fact_type_to_roles = self._parse_fact_type_to_roles()
        self._role_value_constraints = self._parse_role_value_constraints()
        self._internal_uniqueness_constraints, self._external_uniqueness_constraints = self._parse_uniqueness_constraints()
        (self._unique_roles, self._fact_type_to_internal_ucs, self._fact_type_to_complex_ucs,
         self._identifier_fact_type_to_entity_type) = self._process_internal_uniqueness_constraints()
        self._exclusion_constraints = self._parse_exclusion_constraints()
        self._inclusion_constraints = self._parse_inclusion_constraints()
        self._inclusive_role_constraints = self._parse_inclusive_role_constraints()
        self._exclusive_role_constraints = self._parse_exclusive_role_constraints()
        self._mandatory_constraints = self._parse_mandatory_constraints()
        self._ring_constraints = self._parse_ring_constraints()
        self._frequency_constraints = self._parse_frequency_constraints()
        self._value_comparison_constraints = self._parse_value_comparison_constraints()
        self._role_subset_constraints = self._parse_role_subset_constraints()
        self._equality_constraints = self._parse_equality_constraints()
        self._subtype_facts = self._parse_subtype_facts()
        self._sorted_subtype_facts = self._sort_subtype_facts()
        self._fact_type_readings = self._parse_fact_types_reading_orders()
        self._mandatory_roles = self._parse_mandatory_roles()
        self._object_types_to_cardinality_constraints = self._parse_object_types_to_cardinality_constraints()

    def model_name(self):
        return self._model_name

    def entity_types(self) -> dict[str, ORMEntityType]:
        return self._entity_types

    def object_types(self):
        return self._object_types

    def value_types(self):
        return self._value_types

    def roles(self):
        return self._roles

    def role_value_constraints(self):
        return self._role_value_constraints

    def subtype_facts(self):
        return self._subtype_facts

    def external_uniqueness_constraints(self):
        return self._external_uniqueness_constraints

    def value_comparison_constraints(self):
        return self._value_comparison_constraints

    def inclusion_constraints(self):
        return self._inclusion_constraints

    def exclusion_constraints(self):
        return self._exclusion_constraints

    def inclusive_role_constraints(self):
        return self._inclusive_role_constraints

    def exclusive_role_constraints(self):
        return self._exclusive_role_constraints

    def ring_constraints(self):
        return self._ring_constraints

    def frequency_constraints(self):
        return self._frequency_constraints

    def role_subset_constraints(self):
        return self._role_subset_constraints

    def equality_constraints(self):
        return self._equality_constraints

    def cardinality_constraints(self):
        return self._cardinality_constraints

    def value_type_value_constraints(self):
        return self._value_type_value_constraints
    
    def role_cardinality_constraints(self):
        return self._role_cardinality_constraints

    def unique_roles(self):
        return self._unique_roles

    def mandatory_roles(self):
        return self._mandatory_roles

    def fact_type_readings(self):
        return self._fact_type_readings

    def fact_type_to_internal_ucs(self):
        return self._fact_type_to_internal_ucs

    def fact_type_to_complex_ucs(self):
        return self._fact_type_to_complex_ucs

    def fact_type_to_roles(self):
        return self._fact_type_to_roles

    def identifier_fact_type_to_entity_type(self):
        return self._identifier_fact_type_to_entity_type

    def sorted_subtype_facts(self):
        return self._sorted_subtype_facts

    def objects_type_to_cardinality_constraints(self):
        return self._object_types_to_cardinality_constraints

    def _parse_fact_type_to_roles(self):
        fact_type_data = defaultdict(list)
        for role in self._roles.values():
            relationship_name = role.relationship_name
            fact_type_data[relationship_name].append(role)
        return fact_type_data

    def _parse_object_types_to_cardinality_constraints(self):
        object_types_constraints = {}
        for cc in self._cardinality_constraints.values():
            object_type_name = self._object_types[cc.object_type].name
            object_types_constraints[object_type_name] = cc
        return object_types_constraints

    def _parse_model_name(self):
        model_name = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "@Name")
        return model_name if model_name else "ORMModel"

    def _parse_entity_types(self):
        entity_types = {}
        orm_entity_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                            "orm:Objects", "orm:EntityType")
        if orm_entity_types:
            for et in self._single_object_to_list(orm_entity_types):
                id = et["@id"]
                name = et['@Name']
                ref_mode = et.get("@_ReferenceMode") or None
                entity_types[id] = ORMEntityType(id, name, ref_mode)
                self._parse_role_to_player_reference(et)
                self._parse_cardinality_constraint(et)
        return entity_types

    def _parse_value_types(self):
        value_types = {}
        data_types = self._parse_data_types()
        orm_value_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                           "orm:Objects", "orm:ValueType")
        if orm_value_types:
            for vt in self._single_object_to_list(orm_value_types):
                if not vt.get("@IsImplicitBooleanValue"):
                    id = vt["@id"]
                    name = vt["@Name"]
                    data_type = data_types[self._get_nested(vt, "orm:ConceptualDataType", "@ref")]
                    value_types[id] = ORMValueType(id, name, data_type)
                    self._parse_role_to_player_reference(vt)
                    self._parse_cardinality_constraint(vt)
                    self._parse_value_type_value_constraints(vt)
        return value_types

    def _parse_value_type_value_constraints(self, vt):
        vt_id = vt.get("@id")
        value_constraint = self._get_nested(vt,  "orm:ValueRestriction", "orm:ValueConstraint")
        if value_constraint:
            vc_id = value_constraint.get("@id")
            range_values = self._parse_constraint_range(value_constraint)
            self._value_type_value_constraints[vt_id] = ORMValueTypeValueConstraint(vc_id, vt, range_values)

    def _parse_data_types(self):
        data_types = {}
        orm_data_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:DataTypes")
        if orm_data_types:
            for k, v in orm_data_types.items():
                if v is not None:
                    data_types[v["@id"]] = k[4:]
        return data_types

    def _parse_roles(self):
        roles = {}
        ignored_roles = []
        orm_fact_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Facts", "orm:Fact")
        if orm_fact_types:
            for ft in self._single_object_to_list(orm_fact_types):
                relationship_name = ft['@_Name']
                unary_pattern = ft.get("@UnaryPattern", None)  # unary patterns are not supported
                orm_roles = self._get_nested(ft, "orm:FactRoles", "orm:Role")
                if orm_roles:
                    for ro in self._single_object_to_list(orm_roles):
                        role_id = ro["@id"]
                        role_name = ro["@Name"]
                        if (unary_pattern is not None and unary_pattern == "Negation") or role_id not in self._role_to_player.keys():
                            ignored_roles.append(role_id)
                        else:
                            roles[role_id] = ORMRole(role_id, role_name, relationship_name, self._role_to_player[role_id])
                            self._parse_role_cardinality_constraint(ro)
        return roles, ignored_roles

    def _parse_role_to_player_reference(self, object_type):
        roles = self._get_nested(object_type, "orm:PlayedRoles", "orm:Role")
        if roles:
            for role in self._single_object_to_list(roles):
                self._role_to_player[role["@ref"]] = object_type["@id"]

    def _parse_role_value_constraints(self):
        role_value_constraints = {}
        orm_fact_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Facts", "orm:Fact")
        if orm_fact_types:
            for ft in self._single_object_to_list(orm_fact_types):
                roles = self._get_nested(ft, "orm:FactRoles", "orm:Role")
                if roles and isinstance(roles, list):
                    for ro in roles:
                        role_id = ro.get("@id")
                        if role_id not in self._ignored_roles:
                            role_value_constraint = self._get_nested(ro, "orm:ValueRestriction", "orm:RoleValueConstraint")
                            if role_value_constraint:
                                range_values = self._parse_constraint_range(role_value_constraint)
                                role_value_constraints[role_id] = ORMRoleValueConstraint(self._roles[role_id], range_values)
        return role_value_constraints

    def _parse_constraint_range(self, orm_constraint_object):
        values = []
        value_range = self._get_nested(orm_constraint_object, "orm:ValueRanges", "orm:ValueRange")
        if value_range:
            for rvc in self._single_object_to_list(value_range):
                min_value = rvc.get("@MinValue")
                max_value = rvc.get("@MaxValue")
                if min_value and max_value:
                    if min_value == max_value:
                        values.append(min_value)
                    else:
                        values.append(ORMRange(min_value, max_value))
                else:
                    raise ValueError("Range value must have both MinValue and MaxValue defined.")
        return values

    def _parse_inclusive_role_constraints(self):
        constraints = []
        for ec in self._inclusion_constraints.values():
            if ec.roles and set(ec.roles).issubset(self._roles) and all(ro not in self._ignored_roles for ro in ec.roles):
                constraints.append(ORMInclusiveRoleConstraint(ec.id, ec.roles))
        return constraints

    def _parse_exclusive_role_constraints(self):
        constraints = []
        for ec in self._exclusion_constraints.values():
            flat_roles = [role for role_list in ec.roles for role in role_list]
            if ec.roles and set(flat_roles).issubset(self._roles) and all(ro not in self._ignored_roles for ro in flat_roles):
                constraints.append(ORMExclusiveRoleConstraint(ec.id, ec.roles))
        return constraints

    def _parse_subtype_facts(self):
        subtype_of = defaultdict(list)
        for subtype_arrow in self._parse_subtype_arrows():
            subtype_object = self._build_subtype_object(subtype_arrow)
            subtype_of[subtype_object.supertype_name].append(subtype_object)
        return subtype_of

    def _parse_subtype_arrows(self):
        subtype_of = []
        orm_subtype_facts = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                             "orm:Facts", "orm:SubtypeFact")
        if orm_subtype_facts:
            for sft in self._single_object_to_list(orm_subtype_facts):
                subtype_of.append(self._parse_subtype_arrow(sft))
        return subtype_of

    def _parse_subtype_arrow(self, subtype_fact):
        fact_roles = subtype_fact["orm:FactRoles"]
        subtype_id = self._get_nested(fact_roles, "orm:SubtypeMetaRole", "@id")
        subtype = self._get_nested(fact_roles, "orm:SubtypeMetaRole", "orm:RolePlayer", "@ref")
        supertype_id = self._get_nested(fact_roles, "orm:SupertypeMetaRole", "@id")
        supertype = self._get_nested(fact_roles, "orm:SupertypeMetaRole", "orm:RolePlayer", "@ref")
        return ORMSubtypeFact(subtype_id, subtype, supertype_id, supertype)

    def _parse_uniqueness_constraints(self):
        internal = {}
        external = {}
        orm_ucs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:UniquenessConstraint")
        if orm_ucs:
            for uc in self._single_object_to_list(orm_ucs):
                uc_id = uc["@id"]
                pid = uc.get("orm:PreferredIdentifierFor", None)
                identifies = None
                if pid is not None and pid["@ref"] in self._entity_types:
                    identifies = pid["@ref"]
                roles = self._parse_role_sequence(uc)
                if roles:
                    target = internal if uc.get("@IsInternal") == "true" else external
                    target[uc_id] = ORMUniquenessConstraint(uc_id, roles, identifies)
        return internal, external

    def _process_internal_uniqueness_constraints(self):
        unique_roles = []
        fact_type_to_internal_uc = defaultdict(list)
        fact_type_to_complex_uc = defaultdict(list)
        identifier_fact_type_to_entity_type = defaultdict()

        for uc in self._internal_uniqueness_constraints.values():
            if uc.identifies is None and len(uc.roles) == 1:
                unique_roles.extend(uc.roles)

            role_id = uc.roles[0]
            role = self._roles.get(role_id)
            if not role or not role.relationship_name:
                continue

            fact_type = role.relationship_name

            fact_type_to_internal_uc[fact_type].append(uc)

            if uc.identifies is None:
                if len(uc.roles) > 1:
                    fact_type_to_complex_uc[fact_type].append(uc)
            else:
                identifier_fact_type_to_entity_type[fact_type] = self._entity_types[uc.identifies]

        return unique_roles, fact_type_to_internal_uc, fact_type_to_complex_uc, identifier_fact_type_to_entity_type

    def _parse_mandatory_constraints(self):
        mandatory_constraints = {}
        orm_mcs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:MandatoryConstraint")
        if orm_mcs:
            for mc in self._single_object_to_list(orm_mcs):
                mc_id = mc["@id"]
                if mc.get("@IsSimple", None) is not None:
                    roles = self._parse_role_sequence(mc)
                    if roles:
                        mandatory_constraints[mc_id] = ORMMandatoryConstraint(mc_id, roles)
        return mandatory_constraints

    def _parse_value_comparison_constraints(self):
        value_comparison_constraints = {}
        orm_vccs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                    "orm:Constraints", "orm:ValueComparisonConstraint")
        if orm_vccs:
            for vcc in self._single_object_to_list(orm_vccs):
                vcc_id = vcc["@id"]
                vcc_op = vcc["@Operator"]
                roles = self._parse_role_sequence(vcc)
                if roles:
                    constraint = ORMValueComparisonConstraint(vcc_id, ORMValueComparisonOperator(vcc_op), roles)
                    value_comparison_constraints[vcc_id] = constraint
        return value_comparison_constraints

    def _parse_role_subset_constraints(self):
        return self._parse_role_sequence_constraints("orm:SubsetConstraint", ORMRoleSubsetConstraint)

    def _parse_equality_constraints(self):
        return self._parse_role_sequence_constraints("orm:EqualityConstraint", ORMEqualityConstraint)

    def _parse_role_sequence_constraints(self, json_name, constraint_type: type):
        role_sequence_constraints = {}
        orm_constraints = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                    "orm:Constraints", json_name)
        if orm_constraints:
            for role_sequence_constraint in self._single_object_to_list(orm_constraints):
                id = role_sequence_constraint["@id"]
                roles = self._parse_role_sequences(role_sequence_constraint)
                if roles:
                    role_sequence_constraints[id] = constraint_type(id, roles)
        return role_sequence_constraints

    def _parse_inclusion_constraints(self):
        inclusion_constraints = {}
        orm_mcs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:MandatoryConstraint")
        if orm_mcs:
            for mc in self._single_object_to_list(orm_mcs):
                mc_id = mc["@id"]
                mc_name = mc["@Name"]
                if mc_name.startswith("InclusiveOrConstraint"):
                    exclusive = mc.get("orm:ExclusiveOrExclusionConstraint", None) is not None
                    roles = self._parse_role_sequence(mc)
                    if roles:
                        inclusion_constraints[mc_id] = ORMInclusionConstraint(mc_id, roles, exclusive)
        return inclusion_constraints

    def _parse_exclusion_constraints(self):
        exclusion_constraints = {}
        orm_mcs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:ExclusionConstraint")
        if orm_mcs:
            for ec in self._single_object_to_list(orm_mcs):
                ec_id = ec["@id"]
                inclusive = ec.get("orm:ExclusiveOrMandatoryConstraint", None) is not None
                role_sequences = self._parse_role_sequences(ec)
                if role_sequences:
                    exclusion_constraints[ec_id] = ORMExclusionConstraint(ec_id, role_sequences, inclusive)
        return exclusion_constraints

    def _parse_ring_constraints(self):
        ring_constraints = {}
        orm_rcs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:RingConstraint")
        if orm_rcs:
            for rc in self._single_object_to_list(orm_rcs):
                rc_id = rc["@id"]
                rc_type = rc["@Type"]
                roles = self._parse_role_sequence(rc)
                if roles:
                    ring_constraints[rc_id] = ORMRingConstraint(rc_id, roles, self._process_ring_constraint(rc_type))
        return ring_constraints

    def _parse_frequency_constraints(self):
        frequency_constraints = {}
        orm_fcs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:FrequencyConstraint")
        if orm_fcs:
            for fc in self._single_object_to_list(orm_fcs):
                fc_id = fc["@id"]
                min_freq = fc["@MinFrequency"]
                max_freq = fc["@MaxFrequency"]
                roles = self._parse_role_sequence(fc)
                if roles:
                    frequency_constraints[fc_id] = ORMFrequencyConstraint(fc_id, min_freq, max_freq, roles)
        return frequency_constraints

    def _parse_fact_types_reading_orders(self):
        fact_types_readings = defaultdict(list)
        fact_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Facts", "orm:Fact")
        if fact_types:
            for ft in self._single_object_to_list(fact_types):
                relationship_name = (ft['@_Name'])
                reading_orders = self._get_nested(ft, "orm:ReadingOrders", "orm:ReadingOrder")
                if reading_orders:
                    for ro in self._single_object_to_list(reading_orders):
                        reading_order = self._parse_fact_types_reading_order(ro)
                        if reading_order:
                            fact_types_readings[relationship_name].append(reading_order)
        return fact_types_readings

    def _parse_fact_types_reading_order(self, reading_order):
        results = []
        readings = self._get_nested(reading_order, "orm:Readings", "orm:Reading",)
        if readings:
            for rd in self._single_object_to_list(readings):
                expanded_data = self._get_nested(rd,  "orm:ExpandedData")
                if expanded_data:
                    front_text = expanded_data.get("@FrontText", None)
                    if front_text:
                        front_text = front_text.strip()

                    role_texts = self._get_nested(expanded_data,"orm:RoleText")
                    if role_texts:
                        # Step 1: Build roles with whatever RoleText exists
                        index_to_role = {}
                        for rt in self._single_object_to_list(role_texts):
                            index = int(rt["@RoleIndex"])
                            index_to_role[index] = ORMReadingRole(
                                index=index,
                                prefix=rt.get("@PreBoundText", "").strip() or None,
                                postfix=rt.get("@PostBoundText", "").strip() or None,
                                text=rt.get("@FollowingText", "").strip() or None,
                                role=None
                            )

                        # Step 2: Ensure all roles from RoleSequence are represented, even if no RoleText exists
                        roles = self._get_nested(reading_order, "orm:RoleSequence", "orm:Role")
                        if roles:
                            for i, role in enumerate(self._single_object_to_list(roles)):
                                role_id = role.get("@ref")
                                orm_role = self._roles.get(role_id, None)
                                if i in index_to_role:
                                    index_to_role[i].role = orm_role
                                else:
                                    # Fill in with None/defaults if RoleText is missing
                                    index_to_role[i] = ORMReadingRole(
                                        index=i,
                                        prefix=None,
                                        postfix=None,
                                        text=None,
                                        role=orm_role
                                    )
                        results.append(ORMReading(front_text, [index_to_role[i] for i in sorted(index_to_role)]))
        if len(results) > 1:
            warnings.warn("Multiple readings for the same reading order are not supported. Using the first one.")
        return results[0] if results else None

    def _parse_cardinality_constraint_range(self, cardinality_constraint):
        ranges = self._get_nested(cardinality_constraint, "orm:Ranges", "orm:CardinalityRange")
        # At least a range is always present within a cardinality constraint
        orm_ranges = []
        for rg in self._single_object_to_list(ranges):
            range_from = rg.get("@From")
            range_to = rg.get("@To")
            orm_ranges.append(ORMRange(range_from, range_to))
        return orm_ranges

    def _parse_cardinality_constraint(self, object_type):
        orm_cc = self._get_nested(object_type, "orm:CardinalityRestriction", "orm:CardinalityConstraint")
        if orm_cc:
            object_type_id = object_type.get("@id")
            cc_id = orm_cc.get("@id")
            ranges = self._parse_cardinality_constraint_range(orm_cc)
            self._cardinality_constraints[cc_id] = ORMCardinalityConstraint(cc_id, object_type_id, ranges)

    def _parse_role_cardinality_constraint(self, role):
        orm_rcc = self._get_nested(role, "orm:CardinalityRestriction", "orm:UnaryRoleCardinalityConstraint")
        if orm_rcc:
            role_id = role.get("@id")
            cc_id = orm_rcc.get("@id")
            ranges = self._parse_cardinality_constraint_range(orm_rcc)
            self._role_cardinality_constraints[role_id] = ORMRoleCardinalityConstraint(cc_id, role_id, ranges)

    @staticmethod
    def _get_nested(d, *keys):
        for key in keys:
            d = d.get(key)
            if d is None:
                return None
        return d

    def _build_subtype_object(self, subtype_arrow):
        sub_name = self._object_types[subtype_arrow.subtype].name
        sup_name = self._object_types[subtype_arrow.supertype].name
        # Look at exclusion constraints
        for ec_metadata in self._exclusion_constraints.values():
            flat_role_sequence = [role for role_list in ec_metadata.roles for role in role_list]
            if subtype_arrow.supertype_role_id in flat_role_sequence:
                if ec_metadata.inclusive:
                    return ExclusiveInclusiveSubtypeFact(sub_name, sup_name)
                else:
                    return ExclusiveSubtypeFact(sub_name, sup_name)
        # Look at inclusion constraints
        for ic_metadata in self._inclusion_constraints.values():
            if subtype_arrow.supertype_role_id in ic_metadata.roles:
                if ic_metadata.exclusive:
                    return ExclusiveInclusiveSubtypeFact(sub_name, sup_name)
                else:
                    return InclusiveSubtypeFact(sub_name, sup_name)
        # Default case, no constraints found
        return SubtypeFact(sub_name, sup_name)

    def _parse_mandatory_roles(self):
        mandatory_roles = []
        for mc in self._mandatory_constraints.values():
            mandatory_roles.extend(mc.roles)
        return mandatory_roles

    def _parse_role_sequences(self, orm_constraint):
        result = []
        role_sequences = self._get_nested(orm_constraint, "orm:RoleSequences", "orm:RoleSequence")
        if role_sequences:
            for rs in self._single_object_to_list(role_sequences):
                role_sequence = []
                roles = self._get_nested(rs, "orm:Role")
                if roles:
                    for ro in self._single_object_to_list(roles):
                        role_id = ro["@ref"]
                        # Ignore the entire constraint if it involves an ignored role
                        if role_id in self._ignored_roles:
                            return None
                        else:
                            role_sequence.append(role_id)
                    result.append(role_sequence)
        return result

    def _parse_role_sequence(self, orm_constraint):
        roles = self._get_nested(orm_constraint, "orm:RoleSequence", "orm:Role")
        role_sequence = []
        if roles:
            for ro in self._single_object_to_list(roles):
                role_id = ro["@ref"]
                # Ignore the entire constraint if it involves an ignored role
                if role_id in self._ignored_roles:
                    role_sequence = None
                    break
                else:
                    role_sequence.append(role_id)
        return role_sequence

    def _sort_subtype_facts(self):
        # Build a dependency graph to sort subtype facts so that child entity are declared
        # after the parent entity that they extend
        subtype_facts = self._subtype_facts
        edges = []
        nodes = ordered_set()
        sorted_subtype_facts = []
        # Parent depends on child, i.e., parent entity must be declared before child entity
        for parent in subtype_facts:
            nodes.add(parent)
            for child in subtype_facts[parent]:
                target = child.subtype_name
                edges.append((parent, target))
                nodes.add(target)
        for parent in topological_sort(list(nodes), edges):
            sorted_subtype_facts.extend(subtype_facts[parent])
        return sorted_subtype_facts

    @staticmethod
    def _process_ring_constraint(rc_type):
        pattern = '|'.join(r_type.value for r_type in ORMRingType)
        matches = re.finditer(pattern, rc_type)

        ring_types = []
        for match in matches:
            ring_types.append(ORMRingType(match.group(0)))
        return ring_types

    @staticmethod
    def _single_object_to_list(parsed_object):
        return parsed_object if isinstance(parsed_object, list) else [parsed_object]



class RawSource:
    """
    A class representing a raw source.
    """

    _name: str
    _language: str
    _raw_source: str

    def __init__(self, language: str, source_name: str, raw_source: str):
        self._name = f'{source_name}__raw'
        self._language = language
        self._raw_source = raw_source

    @property
    def name(self):
        return self._name

    @property
    def language(self):
        return self._language

    @property
    def raw_source(self):
        return self._raw_source


from io import StringIO
from typing import Optional

import numpy as np
import pandas as pd

import relationalai.early_access.builder as qb
from relationalai.early_access.builder import define, where, builder
from relationalai.early_access.builder.std import decimals

from relationalai.early_access.dsl.bindings.common import BindableColumn, AbstractBindableTable, BindableTable
from relationalai.early_access.dsl.snow.common import CsvColumnMetadata
from relationalai.early_access.dsl.utils import normalize


class BindableCsvColumn(BindableColumn):
    _metadata: CsvColumnMetadata
    _column_basic_type: str

    def __init__(self, metadata: CsvColumnMetadata, table: 'CsvTable', model):
        super().__init__(metadata.name, metadata.datatype, table, model)
        self._metadata = metadata
        self._column_basic_type = "Int64" if metadata.datatype._name == qb.Integer._name else "string"

    def _construct_name(self, col_name: str, col_type: qb.Concept, source: BindableTable):
        """
        Handles the construction of the Relationship name components.

        We use a separate QB relationship per source column, as this allows proper typing.
        """
        short_name = f"{source.physical_name()}_{col_name}"
        madlib = f"{short_name} {{row_id:RowId}} {{{col_type}}}"
        return madlib, short_name

    def __call__(self, *args):
        """
        Allows the column to be called as a function, which is handy for manual data weaving.

        Example:
            row = Integer.ref()
            where(
                {source}.ID(row, id),
                {source}.NAME(row, name),
                person := Person.new(id=id)
            ).define(
                person,
                Person.name(person, name)
            )
        """
        if len(args) != 2:
            raise ValueError(f'Expected 2 arguments passed to a call to BindableColumn, got {len(args)}')
        return qb.Relationship.__call__(self, *args)

    @property
    def metadata(self):
        return self._metadata

    def basic_type(self):
        return self._column_basic_type

    def row_ref(self):
        assert len(self._field_refs) == 2, 'CSV Column relation must have exactly two field references'
        row_ref = self._field_refs[0]
        assert row_ref._thing._name == 'RowId', 'First field must be of type RowId'
        return row_ref

    def decimal_scale(self) -> Optional[int]:
        typ = self.type()
        if decimals.is_decimal(typ):
            return decimals.scale(typ)
        else:
            return None

    def decimal_size(self) -> Optional[int]:
        typ = self.type()
        if decimals.is_decimal(typ):
            return decimals.size(typ)
        else:
            return None

    def __repr__(self):
        return f"CSV:{self._source.physical_name()}.{self.physical_name()}"


class CsvTable(AbstractBindableTable[BindableCsvColumn]):
    _basic_type_schema: dict[str, str]
    _csv_data: list[pd.DataFrame]
    _num_rows: int

    def __init__(self, name: str, schema: dict[str, qb.Concept], model):
        super().__init__(name, model, set())
        self._initialize(schema, model)

    def _initialize(self, schema: dict[str, qb.Concept], model):
        self._csv_data = list()
        self._num_rows = 0
        self._cols = {column_name: BindableCsvColumn(CsvColumnMetadata(column_name, column_type), self, model)
                      for column_name, column_type in schema.items()}
        self._basic_type_schema = {col.metadata.name: col.basic_type() for col in self._cols.values()}

    def __str__(self):
        # returns the name of the table, as well as the columns and their types
        return self.physical_name() + ':\n' + '\n'.join(
            [f' {col.metadata.name} {col.metadata.datatype}' for _, col in self._cols.items()]
        ) + '\n' + '\n'.join(
            [f' {fk.source_columns} -> {fk.target_columns}' for fk in self._foreign_keys]
        )

    @property
    def csv_data(self) -> list[pd.DataFrame]:
        return self._csv_data

    def physical_name(self) -> str:
        return self._table.lower()

    def data(self, csv_data: str):
        csv_df = pd.read_csv(StringIO(normalize(csv_data)), dtype=self._basic_type_schema)
        self._csv_data.append(csv_df)
        CsvSourceModule.generate(self, csv_df, row_offset=self._num_rows)
        # update offset now that we generated the rules
        self._num_rows += len(csv_df)

class CsvSourceModule:

    @staticmethod
    def generate(table: CsvTable, data: pd.DataFrame, row_offset: int = 0):
        for local_index, row in enumerate(data.itertuples(index=False)):
            row_index = row_offset + local_index
            for column_name in data.columns:
                value = getattr(row, column_name)
                if pd.notna(value):
                    column = table.__getattr__(column_name)
                    column_type = column.type()
                    if column_type._name == qb.Date._name:
                        CsvSourceModule._row_to_date_value_rule(column, row_index, value)
                    elif column_type._name == qb.DateTime._name:
                        CsvSourceModule._row_to_date_time_value_rule(column, row_index, value)
                    elif builder.is_decimal(column_type):
                        CsvSourceModule._row_to_decimal_value_rule(column, row_index, value)
                    else:
                        CsvSourceModule._row_to_value_rule(column, row_index, value)

    @staticmethod
    def _row_to_value_rule(column, row, value):
        # if numpy scalar, convert to a native Python type
        if isinstance(value, np.generic):
            value = value.item()
        define(column(row, value))

    @staticmethod
    def _row_to_date_value_rule(column, row, value):
        parse_date = qb.Relationship.builtins['parse_date']
        rez = qb.Date.ref()
        where(parse_date(value, 'Y-m-d', rez)).define(column(row, rez))

    @staticmethod
    def _row_to_date_time_value_rule(column, row, value):
        parse_datetime = qb.Relationship.builtins['parse_datetime']
        rez = qb.DateTime.ref()
        where(parse_datetime(value, 'Y-m-d HH:MM:SS z', rez)).define(column(row, rez))

    @staticmethod
    def _row_to_decimal_value_rule(column, row, value):
        define(column(row, decimals.parse(value, column.type())))



from __future__ import annotations
from collections import defaultdict
from dataclasses import dataclass
import json
import re
import textwrap
from relationalai.clients.util import IdentityParser
from relationalai.early_access.metamodel import types
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.rel.rel_utils import sanitize_identifier
from . import builder as b, annotations as anns
from relationalai import debugging
from relationalai.errors import UnsupportedColumnTypesWarning
from snowflake.snowpark.context import get_active_session

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def schema_dict_to_fields(schema: dict[str, str|b.Concept]) -> list[b.Field]:
    fields = []
    for name, type in schema.items():
        if isinstance(type, str):
            fields.append(b.Field(name=name, type_str=type))
        elif isinstance(type, b.Concept):
            fields.append(b.Field(name=name, type_str=type._name, type=type))
    return fields

#--------------------------------------------------
# Constants
#--------------------------------------------------

# List of supported Snowflake column types. In V0 this comes from the app.get_supported_column_types() function.
# We can't use that function in V1 since it would add a dependency on the app, so we hardcode the list here.
SUPPORTED_SNOWFLAKE_TYPES = [
    'CHAR', 'STRING', 'VARCHAR', 'BINARY', 'NUMBER', 'FLOAT', 'REAL',
    'BOOLEAN', 'DATE', 'FIXED', 'TEXT', 'TIME', 'TIMESTAMP_LTZ',
    'TIMESTAMP_NTZ', 'TIMESTAMP_TZ'
]

#--------------------------------------------------
# Globals
#--------------------------------------------------

_session = None
def get_session():
    global _session
    if not _session:
        try:
            _session = get_active_session()
        except Exception:
            from relationalai import Resources
            from relationalai.clients.snowflake import Resources as SnowflakeResources
            # TODO: we need a better way to handle global config

            # using the resource constructor to differentiate between direct access and
            # regular resource as this impacts the construction of the session.
            resources = Resources()
            assert isinstance(resources, SnowflakeResources), "Can only retrieve session from Snowflake resources"
            _session = resources.get_sf_session()
    return _session

def is_direct_access_enabled():
    from relationalai import Resources
    resources = Resources()
    if hasattr(resources, "is_direct_access_enabled"):
        return resources.is_direct_access_enabled()
    return False


#--------------------------------------------------
# Snowflake sources
#--------------------------------------------------

SFTypes = {
    "TEXT": "String",
    "FIXED": "Number",
    "DATE": "Date",
    "TIMESTAMP": "DateTime",
    "TIMESTAMP_LTZ": "DateTime",
    "TIMESTAMP_TZ": "DateTime",
    "TIMESTAMP_NTZ": "DateTime",
    "FLOAT": "Float",
    "REAL": "Float",
    "BOOLEAN": "Bool",
}

SF_ID_REGEX = re.compile(r'^[A-Za-z_][A-Za-z0-9_$]*$')
def quoted(ident:str):
    if SF_ID_REGEX.match(ident) or ident[0] == '"':
        return ident
    return f'"{ident}"'

@dataclass
class TableInfo:
    source:Table|None
    fields:list[b.Field]
    raw_columns:list[dict]

class SchemaInfo:
    def __init__(self, database:str, schema:str) -> None:
        self.database = database
        self.schema = schema
        self.tables = defaultdict(lambda: TableInfo(None, [], []))
        self.fetched = set()

    def fetch(self):
        session = get_session()
        table_names = [name for name in self.tables.keys() if name.upper() not in self.fetched]
        self.fetched.update([x.upper() for x in table_names])
        name_lookup = {x.upper(): x for x in table_names}
        tables = ", ".join([f"'{x.upper()}', '{x}'" for x in table_names])
        supported_types_list = ", ".join([f"'{t}'" for t in SUPPORTED_SNOWFLAKE_TYPES])

        query = textwrap.dedent(f"""
            begin
                SHOW COLUMNS IN SCHEMA {quoted(self.database)}.{quoted(self.schema)};
                let r resultset := (select "table_name", "column_name", "data_type",
                                    CASE
                                        WHEN PARSE_JSON("data_type"):"type" in ({supported_types_list}) THEN TRUE
                                        ELSE FALSE
                                    END as "supported_type"
                                    from table(result_scan(-1)) as t
                                    where "table_name" in ({tables}));
                return table(r);
            end;
        """)
        with debugging.span("fetch_schema", sql=query):
            columns = session.sql(query).collect()

        # Collect unsupported columns for warning
        unsupported_columns = {}

        for row in columns:
            table_name, column_name, data_type, supported_type = row
            table_name = name_lookup.get(table_name, table_name)

            # Skip columns with unsupported types
            if not supported_type:
                if table_name not in unsupported_columns:
                    unsupported_columns[table_name] = {}
                unsupported_columns[table_name][column_name] = data_type
                continue

            info = self.tables[table_name]
            sf_type_info = json.loads(data_type)
            if sf_type_info.get("type") == "FIXED":
                type_str = self.sf_numeric_to_type_str(column_name, sf_type_info)
            else:
                type_str = SFTypes[sf_type_info.get("type")]
            info.fields.append(b.Field(name=column_name, type_str=type_str))
            info.raw_columns.append(row.as_dict())

        # Show warning for unsupported columns
        if unsupported_columns:
            UnsupportedColumnTypesWarning(unsupported_columns)

    def sf_numeric_to_type_str(self, column_name, sf_type_info):
        """
        Computes the appropriate type to use for this column. This code reflects exactly
        the logic currently used by RAI's CDC implementation to ensure we map to the exact
        same number types.
        """
        if "scale" not in sf_type_info or "precision" not in sf_type_info:
            raise ValueError(f"Invalid definition for column '{column_name}': 'scale' or 'precision' missing")

        precision = sf_type_info.get("precision")
        scale = sf_type_info.get("scale")

        if scale > precision or scale < 0 or scale > 37:
            raise ValueError(f"Invalid numeric scale '{scale}' for column '{column_name}'")

        if scale == 0:
            # Integers (load_csv only supports these two (and not 8/16/32 bit ints)
            # RAI int sizes are 8, 16, 32, 64, or 128; 64 by default
            # SF precision is expressed in base 10 digits, max 38
            # Max positive int value: 127, 32767, 2147483647, 9223372036854775807, 170141183460469231731687303715884105727
            # Base 10 digits fully covered: 2, 4, 9, 18, 38
            bits = types.digits_to_bits(precision)
            return "Int128" if bits == 128 else "Int64"
        else:
            return f"Decimal({precision},{scale})"

class Table():
    _schemas:dict[tuple[str, str], SchemaInfo] = {}
    _used_sources:OrderedSet[Table] = ordered_set()

    def __init__(self, fqn:str, cols:list[str]|None=None, schema:dict[str, str|b.Concept]|None=None) -> None:
        self._fqn = fqn
        parser = IdentityParser(fqn, require_all_parts=True)
        self._database, self._schema, self._table, self._fqn = parser.to_list()
        self._inited = False
        self._concept = b.Concept(self._fqn, extends=[b.Concept.builtins["RowId"]])
        b.Concept.globals[self._fqn.lower()] = self._concept
        self._ref = self._concept.ref("row_id")
        self._cols = {}
        self._col_names = cols
        info = self._schemas.get((self._database, self._schema))
        if not info:
            info = self._schemas[(self._database, self._schema)] = SchemaInfo(self._database, self._schema)
        if schema:
            info.fetched.add(self._table)
            info.tables[self._table] = TableInfo(self, schema_dict_to_fields(schema), [])
        info.tables[self._table].source = self

    def _lazy_init(self):
        if self._inited:
            return
        self._inited = True
        schema_info = self._schemas[(self._database, self._schema)]
        if self._table not in schema_info.fetched:
            schema_info.fetch()
        table_info = schema_info.tables[self._table]
        self._rel = b.Relationship(self._fqn, fields=[b.Field(name="RowId", type_str=self._fqn, type=self._concept)] + table_info.fields)
        self._rel.annotate(anns.external).annotate(anns.from_cdc)

    def __getattr__(self, name: str):
        if name.startswith("_"):
            return super().__getattribute__(name)
        if self._col_names is not None and name not in self._col_names:
            raise ValueError(f"Column {name} not found in {self._fqn}")
        if name.lower() not in self._cols:
            self._cols[name.lower()] = Column(self, name)
        return self._cols[name.lower()]

    def into(self, concept:b.Concept, keys:list[Column]=[]):
        self._lazy_init()
        key_dict = {sanitize_identifier(k._column_name.lower()): k for k in keys}

        # if no keys are defined but the concept has a reference scheme, try to match
        # reference schemes with column fields by name
        if not keys and concept._reference_schemes:
            # do not modify the default argument, create a new array
            keys = []
            sanitized_fields = [sanitize_identifier(n.lower()) for n in self._rel._field_names]
            for scheme in concept._reference_schemes:
                if scheme[0]._name in sanitized_fields:
                    key = Column(self, scheme[0]._name)
                    keys.append(key)
                    key_dict[scheme[0]._name] = key

        # TODO: this is correct, but the rel backend does the wrong thing with it
        # me = concept.new(**key_dict)
        # items = [me]
        # for field in self._rel._fields[1:]:
        #     field_rel = getattr(concept, field.name.lower())
        #     if sanitize_identifier(field.name.lower()) not in key_dict:
        #         items.append(field_rel(me, getattr(self, field.name)))
        # b.where(me).define(*items)

        # this is much less efficient than above

        if keys:
            me = concept.new(**key_dict)
            b.define(me)
        else:
            me = self._rel._field_refs[0]
            b.where(self).define(concept(me))
        for field in self._rel._fields[1:]:
            field_name = sanitize_identifier(field.name.lower())
            if field_name not in key_dict:
                r = b.Relationship(
                    f"{{{concept}}} has {{{field_name}:{field.type_str}}}",
                    parent=concept,
                    short_name=field_name,
                    model=self._rel._model
                )
                setattr(concept, field_name, r)
                relationship = getattr(concept, field_name)
                table_name = getattr(self, field.name)
                update = relationship(me, table_name)
                b.define(update)

    def _to_type(self) -> b.Concept:
        return self._concept

    def _to_keys(self) -> list[b.Ref]:
        self._lazy_init()
        return [self._rel._field_refs[0]]

    def _compile_lookup(self, compiler:b.Compiler, ctx:b.CompilerContext):
        self._lazy_init()
        Table._used_sources.add(self)
        compiler.lookup(self._rel, ctx)
        return compiler.lookup(b.RelationshipFieldRef(None, self._rel, 0), ctx)

class Column(b.Producer):
    def __init__(self, source:Table, column_name:str) -> None:
        super().__init__(None)
        self._column_name = column_name
        self._source = source

    def _to_keys(self) -> list[b.Ref]:
        return self._source._to_keys()

    def _compile_lookup(self, compiler:b.Compiler, ctx:b.CompilerContext):
        compiler.lookup(self._source, ctx)
        for ix, field in enumerate(self._source._rel._field_refs):
            assert field._name is not None
            if field._name.lower() == self._column_name.lower():
                return compiler.lookup(b.RelationshipFieldRef(None, self._source._rel, ix), ctx)
        raise ValueError(f"Column {self._column_name} not found in {self._source._fqn}")



from decimal import Decimal as PyDecimal
import datetime as dt
from typing import Any, Union

from .. import builder as b

_String = Union[b.Producer, str]
_Integer = Union[b.Producer, int]
_Date = Union[b.Producer, dt.date]
_DateTime = Union[b.Producer, dt.datetime]
_Number = Union[b.Producer, float, int, PyDecimal]

def _make_expr(op: str, *args: Any) -> b.Expression:
    return b.Expression(b.Relationship.builtins[op], *args)


from dataclasses import dataclass
from typing import Optional


@dataclass
class OWLClass:
    name: str
    parents: Optional[set[str]] = None
    children: Optional[set[str]] = None
    partition: Optional[set[str]] = None
    disjoint: Optional[set[str]] = None
    mandatory_properties: Optional[set[str]] = None
    inverse_mandatory_properties: Optional[set[str]] = None


@dataclass
class OWLObjectProperty:
    name: str
    domain: list[str]
    range: list[str]
    inverse_property: Optional[str] = None
    functional: Optional[bool] = False
    inverse_functional: Optional[bool] = False


@dataclass
class OWLDatatypeProperty:
    name: str
    domain: list[str]
    range: list[type]
    functional: Optional[bool] = False



import threading
from relationalai.early_access.dsl.orm.models import Model

_model = threading.local()

def init_measure_service(model: Model):
    _model.instance = model

def get_model():
    try:
        return _model.instance
    except AttributeError:
        raise RuntimeError("Measure service is not initialized. Call init_measure_service() before using it.")
    
def get_reasoner():
    return get_model().reasoner()


"""
A simple metamodel for Rel.
"""
from __future__ import annotations
from dataclasses import dataclass, field
from datetime import date, datetime, timezone
from decimal import Decimal as PyDecimal
from io import StringIO
from typing import Optional, Tuple, Union as PyUnion
from relationalai.early_access.metamodel import helpers
from relationalai.early_access.rel import rel_utils
from relationalai.early_access.metamodel.util import OrderedSet, Printer as BasePrinter, ordered_set
import json


#--------------------------------------------------
# Node
#--------------------------------------------------

@dataclass(frozen=True)
class Node:
    def __str__(self):
        return to_string(self)

    @property
    def kind(self):
        return self.__class__.__name__.lower()

#--------------------------------------------------
# Top level program and declarations
#--------------------------------------------------

@dataclass(frozen=True)
class Program(Node):
    # top-level declarations
    declarations: Tuple[PyUnion[Declare, Def, RawSource], ...]

@dataclass(frozen=True)
class Declare(Node):
    """ declare $premise [requires $requires] """
    premise: Expr
    requires: Optional[Expr]
    annotations: Tuple[Annotation, ...] = field(default_factory=tuple)

@dataclass(frozen=True)
class Def(Node):
    """ def $name($params) { $body } """
    name: str
    params: Tuple[PyUnion[Var, Primitive, MetaValue], ...]
    body: Expr
    annotations: Tuple[Annotation, ...] = field(default_factory=tuple)

@dataclass(frozen=True)
class RawSource(Node):
    """ Any string representing Rel code. """
    source: str

#--------------------------------------------------
# Primitives, Annotations
#--------------------------------------------------

Primitive = PyUnion[str, int, float, bool, PyDecimal]

@dataclass(frozen=True)
class Annotation(Node):
    """ @$name($args) """
    name: str
    args: Tuple[PyUnion[Primitive, MetaValue], ...]


#--------------------------------------------------
# Expr
#--------------------------------------------------

@dataclass(frozen=True)
class CompositeExpr(Node):
    pass

Expr = PyUnion[Primitive, CompositeExpr]

@dataclass(frozen=True)
class MetaValue(CompositeExpr):
    """ #$value """
    value: Primitive

@dataclass(frozen=True)
class Var(CompositeExpr):
    """ $name[...] [in $type] """
    name: str
    varargs: bool = False
    type: Optional[str] = None

@dataclass(frozen=True)
class Identifier(CompositeExpr):
    """ $name

    Used to declare a relation by name or to refer to a relation (e.g. Int, String)
    """
    name: str

@dataclass(frozen=True)
class Atom(CompositeExpr):
    """ $expr($args)

    Represents atoms like identifier($name)($args) as well as literal relations
    relations like {(:a, Int, 1)}(x...)
    """
    expr: Expr
    args: Tuple[Expr, ...]

def create_eq(lhs: Expr, rhs: Expr) -> Atom:
    return Atom(Identifier("rel_primitive_eq"), (lhs, rhs))

def atom(name: str, args: Tuple[Expr, ...]):
    """ Helper to create a Atom where expression is an identifier. """
    return Atom(Identifier(name), args)

# Represent the true {()} and false {} relations in Rel
true = Identifier("true")
false = Identifier("false")

@dataclass(frozen=True)
class RelationalAbstraction(CompositeExpr):
    """ ($head): $body """
    head: Tuple[Expr, ...]
    body: Expr

@dataclass(frozen=True)
class And(CompositeExpr):
    """ $body[0] and $body[1] ... and $body[n] """
    body: OrderedSet[Expr]

def create_and(exprs: list[Expr]) -> And:
    """ Create an And expression from this list, pulling out the body of nested Ands. """
    s = ordered_set()
    work = exprs.copy()
    while(work):
        expr = work.pop(0)
        if isinstance(expr, And):
            work.extend(expr.body)
        else:
            s.add(expr)
    return And(s)

@dataclass(frozen=True)
class Or(CompositeExpr):
    """ $body[0] or $body[1] ... or $body[n] """
    body: OrderedSet[Expr]

@dataclass(frozen=True)
class Exists(CompositeExpr):
    """ exists(($vars) | $body ) """
    vars: Tuple[Var, ...]
    body: Expr

@dataclass(frozen=True)
class ForAll(CompositeExpr):
    """ forall(($vars) | $body ) """
    vars: Tuple[Var, ...]
    body: Expr

@dataclass(frozen=True)
class Not(CompositeExpr):
    """ not ( $body ) """
    body: Expr

@dataclass(frozen=True)
class BinaryExpr(CompositeExpr):
    """ $lhs $op $rhs """
    lhs: Expr
    op: str
    rhs: Expr

@dataclass(frozen=True)
class Product(CompositeExpr):
    """ ($body[0] , $body[1] ... , $body[n]) """
    body: Tuple[Expr, ...]

@dataclass(frozen=True)
class Union(CompositeExpr):
    """ {$body[0] ; $body[1] ... ; $body[n]} """
    body: Tuple[Expr, ...]



#--------------------------------------------------
# Printer
#--------------------------------------------------

infix = ["+", "-", "*", "/", "%", "=", "!=", "<", "<=", ">", ">="]

def to_string(node) -> str:
    io = StringIO()
    Printer(io).print_node(node, 0)
    return io.getvalue()

@dataclass(frozen=True)
class Printer(BasePrinter):

    def _join(self, args, sep=', ', indent=0):
        for i, s in enumerate(args):
            if i != 0:
                if indent != 0:
                    self._print(sep.rstrip())
                    self._nl()
                else:
                    self._print(sep)
            self.print_node(s, indent)

    def _print_value(self, value, convert:bool=False):
        if isinstance(value, tuple):
            self._print("(")
            for i, v in enumerate(value):
                if i != 0:
                    self._print(", ")
                self._print_value(v, convert=True)
            self._print(")")
        elif isinstance(value, str):
            self._print(json.dumps(value))
        elif isinstance(value, bool):
            self._print(f"rel_primitive_boolean_{str(value).lower()}")
        elif isinstance(value, int) and convert:
            self._print("::std::common::int[128,")
            self._print(str(value))
            self._print("]")
        else:
            self._print(str(value))

    def print_node(self, node, indent=0) -> None:
        #--------------------------------------------------
        # Top level program and declarations
        #--------------------------------------------------
        if isinstance(node, Program):
            for idx, d in enumerate(node.declarations):
                self.print_node(d, indent)
                # Avoid an extra newline at the end of the file.
                if idx != len(node.declarations) - 1:
                    self._nl()

        elif isinstance(node, Declare):
            for anno in node.annotations:
                self.print_node(anno, indent)
                self._nl()
            self._indent_print(indent, f"declare {node.premise}")
            if node.requires:
                self._print(" requires ")
                self.print_node(node.requires, 0)
            self._nl()

        elif isinstance(node, Def):
            for anno in node.annotations:
                self.print_node(anno, indent)
                self._nl()
            self._indent_print(indent, f"def {helpers.sanitize(node.name)}")
            if node.params:
                self._print("(")
                self._join(node.params)
                self._print("):")
            else:
                self._print(" {")
            self._nl()
            self.print_node(node.body, indent+1)
            self._nl()
            if not node.params:
                self._indent_print(indent, "}")
                self._nl()

        elif isinstance(node, RawSource):
            self._nl()
            self._print(node.source)
            self._nl()

        #--------------------------------------------------
        # Primitives, Annotations
        #--------------------------------------------------
        elif node is None:
            self._indent_print(indent, "")
            self._print("missing")

        elif isinstance(node, (str, int, float, bool, PyDecimal, tuple)):
            self._indent_print(indent, "")
            self._print_value(node)

        elif isinstance(node, date) and not isinstance(node, datetime):
            self._indent_print(indent, node.isoformat())

        elif isinstance(node, datetime):
            if node.tzinfo is None:
                node = node.replace(tzinfo=timezone.utc)

            self._indent_print(indent, node.astimezone(timezone.utc).isoformat())

        elif isinstance(node, Annotation):
            self._indent_print(indent, f"@{node.name}")
            if node.args:
                self._print("(")
                self._join(node.args)
                self._print(")")


        #--------------------------------------------------
        # Expr
        #--------------------------------------------------

        elif isinstance(node, MetaValue):
            if isinstance(node.value, str):
                self._print(":")
            else:
                self._print("#")
            self._print_value(node.value)

        elif isinstance(node, Var):
            self._indent_print(indent, f"{node.name}")
            if node.varargs:
                self._print("...")
            if node.type and node.type != "Any":
                self._print(f" in {node.type if node.type != 'Enum' else 'UInt128'}")

        elif isinstance(node, Identifier):
            self._indent_print(indent, helpers.sanitize(node.name))

        elif isinstance(node, Atom):
            if isinstance(node.expr, Identifier) and node.expr.name in infix:
                # deal with the 3 kinds of infix operators
                if len(node.args) == 1:
                    self.print_node(node.expr, indent)
                    self.print_node(node.args[0])
                elif len(node.args) == 2:
                    self.print_node(node.args[0], indent)
                    self._print(" ")
                    self.print_node(node.expr)
                    self._print(" ")
                    self.print_node(node.args[1])
                elif len(node.args) == 3:
                    self.print_node(node.args[2], indent)
                    self._print(" = ")
                    self.print_node(node.args[0])
                    self._print(" ")
                    self.print_node(node.expr)
                    self._print(" ")
                    self.print_node(node.args[1])
                else:
                    raise NotImplementedError(f"emit_action: {node}")
            else:
                is_higher_order = False
                is_like = False
                if isinstance(node.expr, Identifier):
                    is_higher_order = node.expr.name in rel_utils.HIGHER_ORDER
                    is_like = node.expr.name == "::std::common::like_match"

                    self.print_node(node.expr, indent)
                elif isinstance(node.expr, Union):
                    self.print_node(node.expr, indent)
                else:
                    self._indent_print(indent, "{")
                    self.print_node(node.expr, indent)
                    self._print("}")

                self._print("(")
                if is_higher_order:
                    self._nl()
                    self._join(node.args, indent=indent+1)
                    self._nl()
                    self._indent_print(indent, ")")
                elif is_like:
                    self._print("raw\"")
                    self._print(node.args[1])
                    self._print("\", ")
                    self.print_node(node.args[0])
                    self._print(")")
                else:
                    self._join(node.args)
                    self._print(")")

        elif isinstance(node, RelationalAbstraction):
            if node.head:
                self._indent_print(indent, "{(")
                self._join(node.head)
                self._print("):")
                if self._is_simple_body(node.body):
                    self._print(" ")
                    self.print_node(node.body)
                    self._print("}")
                else:
                    self._nl()
                    self.print_node(node.body, indent+1)
                    self._nl()
                    self._indent_print(indent, "}")
            else:
                self.print_node(node.body, indent)

        elif isinstance(node, And):
            self._join(node.body, " and ", indent)

        elif isinstance(node, Or):
            self._indent_print(indent, "(")
            self._join(node.body, " or ")
            self._print(")")

        elif isinstance(node, Exists):
            self._indent_print(indent, "exists((")
            self._join(node.vars)
            self._print(") |")
            if self._is_simple_body(node.body):
                self._print(" ")
                self.print_node(node.body)
                self._print(")")
            else:
                self._nl()
                self.print_node(node.body, indent+1)
                self._nl()
                self._indent_print(indent, ")")

        elif isinstance(node, ForAll):
            self._indent_print(indent, "forall((")
            self._join(node.vars)
            self._print(") |")
            if self._is_simple_body(node.body):
                self._print(" ")
                self.print_node(node.body)
                self._print(")")
            else:
                self._nl()
                self.print_node(node.body, indent+1)
                self._nl()
                self._indent_print(indent, ")")

        elif isinstance(node, Not):
            self._indent_print(indent, "not (")
            if self._is_simple_body(node.body):
                self._print(" ")
                self.print_node(node.body)
                self._print(" )")
            else:
                self._nl()
                self.print_node(node.body, indent+1)
                self._nl()
                self._indent_print(indent, ")")

        elif isinstance(node, BinaryExpr):
            self.print_node(node.lhs, indent)
            self._print(f" {node.op} ")
            self.print_node(node.rhs)

        elif isinstance(node, Product):
            self._indent_print(indent, "(")
            self._join(node.body, ", ")
            self._print(")")

        elif isinstance(node, Union):
            if all(isinstance(b, Atom) for b in node.body):
                self._join(node.body, " or ", indent)
            else:
                self._indent_print(indent, "{")
                self._nl()
                self._join(node.body, " ; ", indent+1)
                self._nl()
                self._indent_print(indent, "}")

        else:
            raise Exception(f"Missing implementation in Rel printer: {type(node)}")

    def _is_simple_body(self, node: Expr) -> bool:
        if isinstance(node, (Var, Primitive, MetaValue, Identifier)):
            return True
        elif isinstance(node, Atom):
            return self._is_simple_body(node.expr) and all(self._is_simple_body(a) for a in node.args)
        elif isinstance(node, Exists):
            return self._is_simple_body(node.body)
        elif isinstance(node, ForAll):
            return self._is_simple_body(node.body)
        elif isinstance(node, Not):
            return self._is_simple_body(node.body)
        elif isinstance(node, BinaryExpr):
            return self._is_simple_body(node.lhs) and self._is_simple_body(node.rhs)
        return False



import logging

from typing import Any

#--------------------------------------------------
# A custom logging handler to grab compilation
# and result information from programs that run
#--------------------------------------------------

class Capturer(logging.Handler):
    def __init__(self):
        super().__init__()
        self.starts = {}
        self.model = []
        self.queries = []
        self.results = []
        self.types = []

    def emit(self, record):
        obj:Any = record.args
        if not obj:
            return

        if obj["event"] == "span_start":
            span = obj["span"]
            self.starts[str(span.id)] = span
        elif obj["event"] == "span_end":
            start = self.starts.pop(obj["id"], None)
            if start:
                props = start.attrs.copy()
                props["type"] = start.type
                props.update(obj.get("end_attrs", {}))
                props["elapsed"] = obj["elapsed"]
                self.done(props)

    def done(self, info):
        if info.get("type") == "compile" and info.get("compile_type") == "model":
            self.model.append(info)
        if info.get("type") == "compile" and info.get("compile_type") == "query":
            self.queries.append(info)
        if info.get("type") == "query":
            self.results.append(info)
        if info.get("type") == "type.propagate":
            self.types.append(info)


from typing import Union

from relationalai.early_access.dsl.bindings.common import Binding, IdentifierConceptBinding, RoleBinding, \
    SubtypeConceptBinding, BindableTable, BindableAttribute, ReferentConceptBinding
from relationalai.early_access.dsl.codegen.common import BoundIdentifierConstraint, BoundRelationship
from relationalai.early_access.dsl.orm.relationships import Role, Relationship
from relationalai.early_access.metamodel.util import OrderedSet

import relationalai.early_access.builder as qb

class Binder:
    def __init__(self, model):
        self._model = model
        self._reasoner = self._model.reasoner()

        # bindings classification
        self._identifier_bindings: OrderedSet[IdentifierConceptBinding] = OrderedSet()
        self._subtype_bindings: OrderedSet[Union[IdentifierConceptBinding, ReferentConceptBinding]] = OrderedSet()
        self._subtype_concept_bindings: OrderedSet[SubtypeConceptBinding] = OrderedSet()
        self._referent_concept_bindings: OrderedSet[ReferentConceptBinding] = OrderedSet()
        self._value_type_bindings: OrderedSet[Binding] = OrderedSet()
        self._entity_type_bindings: OrderedSet[Binding] = OrderedSet()
        self._role_bindings: dict[Role, OrderedSet[Binding]] = {}
        self._bound_identifier_constraints = OrderedSet[BoundIdentifierConstraint]()
        self._bound_relationships = OrderedSet[BoundRelationship]()

        self._analyze()

    def value_type_bindings(self) -> OrderedSet[Binding]:
        """
        Returns the value type bindings.
        """
        return self._value_type_bindings

    def is_value_type_binding(self, binding: Binding) -> bool:
        """
        Returns True if the binding is a value type binding.
        """
        return binding in self._value_type_bindings

    def constructor_bindings(self) -> OrderedSet[IdentifierConceptBinding]:
        """
        Returns the identifier bindings that construct entity types.
        """
        return self._identifier_bindings

    def is_constructor_binding(self, binding: Binding) -> bool:
        """
        Returns True if the binding is an IdentifierBinding.
        """
        return isinstance(binding, IdentifierConceptBinding) and binding in self._identifier_bindings

    @staticmethod
    def is_filtering_binding(binding: Binding) -> bool:
        """
        Returns True if the binding is a SubtypeConceptBinding or has filter_by.
        """
        return isinstance(binding, SubtypeConceptBinding) or binding.filter_by is not None

    def subtype_ctor_bindings(self) -> OrderedSet[Union[IdentifierConceptBinding, ReferentConceptBinding]]:
        """
        Returns the identifier or referent subtype bindings. Those are only possible for subtypes that define
        their own reference scheme.

        They are different from SubtypeConceptBinding, which is a subtype binding with a filter.
        """
        return self._subtype_bindings

    def subtype_filtering_bindings(self) -> OrderedSet[SubtypeConceptBinding]:
        """
        Returns the subtype concept bindings. These are bindings that are used to derive subtypes from a parent type,
        and they are different from IdentifierConceptBinding or ReferentConceptBinding, which are used to construct
        entities or refer to them.
        """
        return self._subtype_concept_bindings

    def referent_concept_bindings(self) -> OrderedSet[ReferentConceptBinding]:
        """
        Returns the referent concept bindings.
        """
        return self._referent_concept_bindings

    def lookup_binding_role(self, binding: Binding):
        if isinstance(binding, RoleBinding):
            # simple case: use the Role from the binding
            role = binding.role
        elif isinstance(binding, (IdentifierConceptBinding, SubtypeConceptBinding, ReferentConceptBinding)):
            # either an IdentifierBinding or a SubtypeBinding, look up the role in the IdentifierRelationship
            concept = binding.entity_type
            if self._reasoner.has_own_ref_scheme(concept):
                role = self._reasoner.own_ref_scheme_role(concept)
            else:
                # this is a subtype binding...
                role = self._reasoner.subtype_unique_ref_scheme_role(concept)
        else:
            raise ValueError(f"Unknown binding type: {type(binding)}")
        return role

    def lookup_role_bindings(self, role: Role, source: BindableTable) -> OrderedSet[Binding]:
        """
        Look up bindings for the given Role and source table.
        Returns an OrderedSet of bindings.
        """
        if role not in self._role_bindings:
            raise ValueError(f"No bindings found for role {role} in table {source}")
        bindings = [binding for binding in self._role_bindings[role] if binding.column.table == source]
        return OrderedSet().update(bindings)

    def is_role_bound(self, role: Role, source: BindableTable) -> bool:
        """
        Check if the given Role is bound in the source table.
        """
        return role in self._role_bindings and bool(self.lookup_role_bindings(role, source))

    def lookup_constructor_binding(self, concept: qb.Concept, column: BindableAttribute):
        """
        Look up a constructor binding for the given Concept and source table.
        """
        if concept._is_primitive():
            raise ValueError(f"Cannot look up entity constructor binding for a value type: {concept}")
        candidates = list(self._identifier_bindings) + list(self._referent_concept_bindings)
        for binding in candidates:
            candidate_role = self.lookup_binding_role(binding)
            sibling = candidate_role.sibling()
            assert sibling is not None
            if binding.column is column and sibling.player() is concept:
                return binding
        raise ValueError(f"Cannot look up entity constructor binding for {concept} from {column}")

    def lookup_constructor_binding_by_source(self, concept: qb.Concept, source: BindableTable):
        """
        Look up a constructor binding for the given Concept and source table.
        """
        if concept._is_primitive():
            raise ValueError(f"Cannot look up entity constructor binding for a value type: {concept}")
        for binding in self._identifier_bindings:
            sibling_role = self.lookup_binding_role(binding).sibling()
            assert sibling_role is not None
            if binding.column.table == source and sibling_role.player() is concept:
                return binding
        raise ValueError(f"Cannot look up entity constructor binding for {concept} from {source}")

    def lookup_subtype_reference_binding(self, binding: Union[IdentifierConceptBinding, SubtypeConceptBinding]):
        """
        Look up a constructor binding referenced by the subtype binding.
        """
        ctor_role = self._reasoner.subtype_unique_ref_scheme_role(binding.entity_type)
        sibling_role = ctor_role.sibling()
        assert sibling_role is not None
        parent_type = sibling_role.player()
        return self.lookup_constructor_binding(parent_type, binding.column)

    def lookup_subtype_reference_binding_by_source(self, binding: Union[IdentifierConceptBinding, SubtypeConceptBinding]):
        """
        Look up a constructor binding referenced by the subtype binding.
        """
        ctor_role = self._reasoner.subtype_unique_ref_scheme_role(binding.entity_type)
        sibling_role = ctor_role.sibling()
        assert sibling_role is not None
        parent_type = sibling_role.player()
        return self.lookup_constructor_binding_by_source(parent_type, binding.column.table)

    def lookup_bound_identifier_constraint_by_source(self, concept: qb.Concept, source: BindableTable):
        """
        Look up a bound identifier constraint for the given Concept and source table.
        """
        identifier_constraints = self._reasoner.ref_schemes_of(concept)
        bound_constraints = OrderedSet()
        for bound_constraint in self._bound_identifier_constraints:
            if bound_constraint.constraint in identifier_constraints and bound_constraint.table == source:
                bound_constraints.add(bound_constraint)
        #=
        # Note: this is a bit naive right now, it should take into account whether the concept is being:
        #  * constructed (we need ALL) or;
        #  * referenced (we need exactly one).
        #
        # To be addressed in https://relationalai.atlassian.net/browse/RAI-39757.
        #=
        if not (bound_constraints and all(b.constraint in identifier_constraints for b in bound_constraints)):
            raise ValueError(
                f"Cannot look up bound identifier constraints for {concept} from {source}, "
                f"no bound constraints found"
            )
        return bound_constraints

    def bound_identifier_constraints(self):
        """
        Returns the bound identifier constraints.
        """
        return self._bound_identifier_constraints

    def lookup_bound_identifier_constraint_by_column(self, concept: qb.Concept, column: BindableAttribute):
        """
        Look up a bound identifier constraint for the given column and concept.
        """
        for constraint in self._bound_identifier_constraints:
            if constraint.concept is concept and constraint.table == column.table:
                for _, binding in constraint.role_bindings.items():
                    if binding.column is column:
                        return constraint
        raise ValueError(f"Cannot look up bound identifier constraint for {concept} from {column.table}")

    def bound_relationships(self) -> OrderedSet[BoundRelationship]:
        """
        Returns the bound relationships.
        """
        return self._bound_relationships

    #=
    # Analysis based on the bindings in the model.
    #=

    def _analyze(self):
        self._classify_bindings()
        self._classify_constraints()
        self._classify_relationships()

    #=
    # Look through the bindings and classify them.
    #=

    def _classify_bindings(self):
        """
        Analyze the bindings in the model and update the reasoner.
        """
        for binding in self._model._bindings:
            self._classify_binding(binding)

    def _classify_binding(self, binding: Binding):
        if isinstance(binding, Union[IdentifierConceptBinding, ReferentConceptBinding]):
            self._process_concept_binding(binding)
        elif isinstance(binding, RoleBinding):
            self._process_role_binding(binding)
        elif isinstance(binding, SubtypeConceptBinding):
            self._process_subtype_concept_binding(binding)
        else:
            raise ValueError(f"Unknown binding type: {type(binding)}")

    def _process_concept_binding(self, binding: Union[IdentifierConceptBinding, ReferentConceptBinding]):
        """
        ConceptBinding could either be an IdentifierBinding or a SubtypeBinding.
        It depends on whether the entity type has a reference scheme or not, if so
        it is an IdentifierBinding, otherwise it is a SubtypeBinding.
        """
        concept = binding.entity_type
        is_referent_binding = isinstance(binding, ReferentConceptBinding)
        if self._reasoner.has_own_ref_scheme(concept):
            # If the concept has its own reference scheme, it is an IdentifierBinding,
            # which could be a referent (if there's a FK to another table) or constructor binding.
            # ReferentConceptBinding is a special case of ConceptBinding, where users tell explicitly
            # that the binding is a referent.
            if not is_referent_binding:
                self._identifier_bindings.add(binding)
            constructing_role = self._reasoner.own_ref_scheme_role(concept)
        else:
            self._subtype_bindings.add(binding)
            constructing_role = self._reasoner.subtype_unique_ref_scheme_role(concept)
        if is_referent_binding:
            self._referent_concept_bindings.add(binding)
        self._role_bindings.setdefault(constructing_role, OrderedSet()).add(binding)
        ref_mode_concept = constructing_role.player()
        self._classify_binding_by_type(binding, ref_mode_concept)

    def _process_role_binding(self, binding: RoleBinding):
        """
        RoleBinding is a simple binding to a Role in a Relationship. The player of the role could
        be either a ValueType or an EntityType.
        """
        self._classify_binding_by_type(binding, binding.role.player())
        self._role_bindings.setdefault(binding.role, OrderedSet()).add(binding)

    def _process_subtype_concept_binding(self, binding: SubtypeConceptBinding):
        """
        FilteringSubtypeBinding is a subtype binding with a filter.

        It is neither an entity nor a value type binding. The filter may be a literal or a value from a
        RoleValueConstraint.
        """
        self._subtype_concept_bindings.add(binding)

    def _classify_binding_by_type(self, binding: Binding, player: qb.Concept):
        if isinstance(binding, SubtypeConceptBinding):
            # SubtypeConceptBinding is neither an entity nor a value type binding.
            return
        if player._is_primitive():
            self._value_type_bindings.add(binding)
        else:
            self._entity_type_bindings.add(binding)

    def _classify_constraints(self):
        for concept, identifier_constraints in self._reasoner.concept_identifiers().items():
            for constraint in identifier_constraints:
                roles = constraint.roles()
                role_bindings_by_table = self._collect_table_role_bindings(roles)
                for table, bindings in role_bindings_by_table.items():
                    if len(bindings) < len(roles):
                        raise ValueError(f"Not all roles are bound in {table.physical_name()}: {bindings} for {roles}")
                    bound_constraint = BoundIdentifierConstraint(constraint, table, bindings)
                    self._bound_identifier_constraints.add(bound_constraint)

    def _collect_table_role_bindings(self, roles):
        role_bindings_by_table = {}
        for role in roles:
            bindings = self._role_bindings.get(role)
            if not bindings:
                continue
            for binding in bindings:
                table = binding.column.table
                role_bindings_by_table.setdefault(table, {}).setdefault(role, []).append(binding)
        return role_bindings_by_table

    def _classify_relationships(self):
        bound_relationship_bindings = {}
        for role, bindings in self._role_bindings.items():
            relationship = role._part_of()
            assert isinstance(relationship, Relationship)
            for binding in bindings:
                table = binding.column.table
                bound_relationship = BoundRelationship(relationship, table)
                bound_relationship_bindings.setdefault(bound_relationship, []).append(binding)

        for bound_relationship, bindings in bound_relationship_bindings.items():
            bound_relationship.bindings.extend(bindings)
            self._validate_bound_relationship(bound_relationship)

    def _validate_bound_relationship(self, bound_relationship: BoundRelationship):
        relationship = bound_relationship.relationship
        arity = relationship._arity()

        if arity == 1 and bound_relationship.bindings:
            self._bound_relationships.add(bound_relationship)
        elif arity == 2:
            self._validate_binary_bound_relationship(bound_relationship)
        elif arity > 2:
            self._validate_nary_bound_relationship(bound_relationship)

    def _validate_binary_bound_relationship(self, bound_relationship: BoundRelationship):
        relationship = bound_relationship.relationship
        if self._reasoner.is_identifier_relationship(relationship):
            # ignore as this is handled by constructor entity maps
            return
        #=
        # Typically, at least one of the roles in a binary relationship must be bound.
        # In such cases, we attempt to infer the entity map for the unbound role, but
        # any value type role must be bound. If unable to infer the entity map, an error
        # is raised.
        #=
        key_role, value_role = self._identify_binary_relationship_roles(relationship)
        # the value role may infer if it's an entity type and the key role is bound
        value_role_may_infer =\
            not value_role.player()._is_primitive() and self.is_role_bound(key_role, bound_relationship.table)
        self._assert_role_is_bound(value_role, bound_relationship.table, may_infer=value_role_may_infer)
        self._assert_role_is_bound(key_role, bound_relationship.table, may_infer=True)
        self._bound_relationships.add(bound_relationship)

    def _identify_binary_relationship_roles(self, relationship: Relationship) -> tuple[Role, Role]:
        roles = relationship._roles()
        key_role, value_role = None, None
        for role in roles:
            if role.player()._is_primitive():
                if value_role is not None:
                    raise ValueError(f"Binary relationship {relationship} has multiple value type roles: "
                                     f"{value_role} and {role}")
                value_role = role
            elif key_role is None and not self._reasoner.is_constructing_role(role):
                key_role = role
            else:
                value_role = role
        if not isinstance(key_role, Role) and not isinstance(value_role, Role):
            raise ValueError(f'Cannot identify roles in a binary relationship {relationship}')
        return key_role, value_role  # pyright: ignore[reportReturnType]

    def _assert_role_is_bound(self, role: Role, source: BindableTable, may_infer=False):
        """
        Asserts that the role is bound in the context of a tabular source.
        If may_infer is True, it allows for inferring the entity map, only for roles played by
        an entity type.
        """
        if not self._role_bindings.get(role):
            # value type roles must be bound, entity type roles may infer
            may_infer = False if role.player()._is_primitive() else may_infer
            if may_infer and self.lookup_bound_identifier_constraint_by_source(role.player(), source):
                # infer the binding from the relationship
                return
            raise ValueError(f"Role {role} in relationship {role._part_of()} is not bound")

    def _validate_nary_bound_relationship(self, bound_relationship: BoundRelationship):
        relationship = bound_relationship.relationship
        roles = relationship._roles()
        bound_roles = OrderedSet()
        for binding in bound_relationship.bindings:
            if not isinstance(binding, RoleBinding):
                raise ValueError(f"Only RoleBindings allowed in N-ary relationships, got: {type(binding)}")
            bound_roles.add(binding.role)

        has_entity_type_player, is_bound = False, True
        for role in roles:
            if not (role in bound_roles or
                    self.lookup_bound_identifier_constraint_by_source(role.player(), bound_relationship.table)):
                is_bound = False
                break
            if not role.player()._is_primitive():
                has_entity_type_player = True

        if is_bound and has_entity_type_player:
            self._bound_relationships.add(bound_relationship)
        elif not is_bound:
            raise ValueError(f"N-ary relationship {relationship} is not fully bound: {bound_roles} vs {roles}")
        else:
            raise ValueError(f"N-ary relationship {relationship} must have at least one entity type player, "
                             f"but all players are value types: {roles}")



"""
    Functions to simplify the creation of IR nodes using some common variations.
"""
from __future__ import annotations
from typing import Any, Tuple, Optional, Sequence as PySequence, Union
import datetime
import decimal


from relationalai.early_access.metamodel import ir, types
from relationalai.early_access.metamodel.util import OrderedSet, FrozenOrderedSet, ordered_set, frozen

import pandas as pd

#-------------------------------------------------
# Public Types - Model
#-------------------------------------------------

def model(
        engines: OrderedSet[ir.Engine],
        relations: OrderedSet[ir.Relation],
        types: OrderedSet[ir.Type],
        root: ir.Task,
        annos: PySequence[ir.Annotation]=[]):
    return ir.Model(
        engines.frozen(),
        relations.frozen(),
        types.frozen(),
        root,
        FrozenOrderedSet(annos)
    )

def compute_model(root: ir.Task) -> ir.Model:
    from relationalai.early_access.metamodel.visitor import collect_by_type
    return model(
        collect_by_type(ir.Engine, root),
        collect_by_type(ir.Relation, root),
        _collect_reachable_types(collect_by_type(ir.Type, root)),
        root
    )

def _collect_reachable_types(type_set: OrderedSet[ir.Type]) -> OrderedSet[ir.Type]:
    """ Collect all types reachable from these types + builtins, including super types, element types, etc. """
    # add all built-in types
    type_set.update(types.builtin_types)

    # add super types and element types of collections
    to_visit = type_set.get_list().copy()
    while to_visit:
        t = to_visit.pop()
        type_set.add(t)
        if isinstance(t, ir.ScalarType):
            to_visit.extend(t.super_types)
        elif isinstance(t, (ir.ListType)):
            to_visit.append(t.element_type)
        elif isinstance(t, (ir.UnionType, ir.TupleType)):
            to_visit.extend(t.types)
    return type_set

#-------------------------------------------------
# Public Types - Engine
#-------------------------------------------------

def capability(name: str):
    return ir.Capability(name)

def engine(name: str, platform: str, relations:OrderedSet[ir.Relation], info: Any=None, capabilities: OrderedSet[ir.Capability]=ordered_set(), annos: PySequence[ir.Annotation]=[]):
    return ir.Engine(name, platform, info, capabilities.frozen(), relations.frozen(), FrozenOrderedSet(annos))

#-------------------------------------------------
# Public Types - Data
#-------------------------------------------------

def scalar_type(name: str, super_types:list[ir.ScalarType]=[], annos: PySequence[ir.Annotation]=[]):
    return ir.ScalarType(name, FrozenOrderedSet(super_types), FrozenOrderedSet(annos))

def decimal_type(name: str, precision: int, scale: int, annos: PySequence[ir.Annotation]=[]):
    return ir.DecimalType(name, frozen(), FrozenOrderedSet(annos), precision, scale)

def list_type(element_type: ir.Type):
    return ir.ListType(element_type)

def union_type(types: list[ir.Type]):
    return ir.UnionType(FrozenOrderedSet(types))

def field(name: str, type: ir.Type, input: bool=False):
    return ir.Field(name, type, input)

def input_field(name:str, type: ir.Type):
    return ir.Field(name, type, True)


# property helpers
def relation(name: str, fields: list[ir.Field], requires: OrderedSet[ir.Capability]=ordered_set(), annos: PySequence[ir.Annotation]=[], overloads: PySequence[ir.Relation]=[]):
    return ir.Relation(name, tuple(fields), requires.frozen(), FrozenOrderedSet(annos), FrozenOrderedSet(overloads))

def entity(type: ir.ScalarType):
    return relation(type.name, [field("id", type)])

def property(name: str, key_name: str, key_type: ir.Type, value_name:str, value_type: ir.Type):
    return relation(
        name,
        [field(key_name, key_type), field(value_name, value_type)]
    )

#-------------------------------------------------
# Public Types - Tasks
#-------------------------------------------------

def success():
    return ir.Task(None)

#
# Task composition
#

def logical(body: PySequence[ir.Task], hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Logical(engine, tuple(hoisted), tuple(body), FrozenOrderedSet(annos))

def sequence(tasks: PySequence[ir.Task], hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Sequence(engine, tuple(hoisted), tuple(tasks), FrozenOrderedSet(annos))

def union(tasks: PySequence[ir.Task], hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Union(engine, tuple(hoisted), tuple(tasks), FrozenOrderedSet(annos))

def match(tasks: PySequence[ir.Task], hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Match(engine, tuple(hoisted), tuple(tasks), FrozenOrderedSet(annos))

def until(check: ir.Task, body: ir.Task, hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Until(engine, tuple(hoisted), check, body, FrozenOrderedSet(annos))

def wait(check: ir.Task, hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Wait(engine, tuple(hoisted), check, FrozenOrderedSet(annos))

#
# Constraints
#

def check(check:ir.Task, error:Optional[ir.Task], annos: PySequence[ir.Annotation]=[]):
    return ir.Check(check, error, FrozenOrderedSet(annos))

def require(domain:ir.Task, checks:PySequence[ir.Check], engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Require(engine, domain, tuple(checks), FrozenOrderedSet(annos))

#
# Relational Operations
#

def var(name: str, type: Union[ir.Type, None] = None) -> ir.Var:
    if type is None:
        return ir.Var(types.Any, name)
    else:
        return ir.Var(type, name)

def default(var: ir.Var, value: ir.Value) -> ir.Default:
    return ir.Default(var, value)

def wild(type: ir.Type) -> ir.Var:
    return var("_", type)

def literal(value: Any, type: Union[ir.Type, None] = None) -> ir.Value:
    """ Create a Literal with this type. If type is not present, attempt to figure out the
    appropriate type. See: `l`. """
    if type is None:
        return lit(value)
    else:
        return ir.Literal(type, value)

def data(df: pd.DataFrame, vars: PySequence[ir.Var], engine:Optional[ir.Engine]=None) -> ir.Data:
    """ Create a DataFrame literal. """
    return ir.Data(engine, df, tuple(vars))

def lit(value: Any) -> ir.Value:
    """ Ensure this value is an appropriate ir.Value. This function wraps common python values
     in ir.Literals, and supports lists of values. """
    if isinstance(value, ir.Literal):
        return value
    elif isinstance(value, str):
        return ir.Literal(types.String, value)
    elif isinstance(value, int):
        return ir.Literal(types.Int128, value)
    elif isinstance(value, float):
        return ir.Literal(types.Float, value)
    elif isinstance(value, bool):
        return ir.Literal(types.Bool, value)
    elif isinstance(value, decimal.Decimal):
        return ir.Literal(types.Decimal, value)
    elif isinstance(value, datetime.date):
        return ir.Literal(types.Date, value)
    elif isinstance(value, datetime.datetime):
        return ir.Literal(types.DateTime, value)
    elif isinstance(value, list):
        return tuple([lit(v) for v in value])
    else:
        raise NotImplementedError(f"literal value: {value} of type {type(value)}")

def annotation(relation: ir.Relation, args: PySequence[ir.Value]) -> ir.Annotation:
    return ir.Annotation(relation, tuple(args))

def derive(relation: ir.Relation, args: PySequence[ir.Value], annos: PySequence[ir.Annotation]=[], engine: Optional[ir.Engine]=None):
    return ir.Update(engine, relation, tuple(args), ir.Effect.derive, FrozenOrderedSet(annos))

def insert(relation: ir.Relation, args: PySequence[ir.Value], annos: PySequence[ir.Annotation]=[], engine: Optional[ir.Engine]=None):
    return ir.Update(engine, relation, tuple(args), ir.Effect.insert, FrozenOrderedSet(annos))

def delete(relation: ir.Relation, args: PySequence[ir.Value], annos: PySequence[ir.Annotation]=[], engine: Optional[ir.Engine]=None):
    return ir.Update(engine, relation, tuple(args), ir.Effect.delete, FrozenOrderedSet(annos))

def lookup(relation: ir.Relation, args: PySequence[ir.Value], engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Lookup(engine, relation, tuple(args), FrozenOrderedSet(annos))

def update(relation: ir.Relation, args: PySequence[ir.Value], effect: ir.Effect, annos: PySequence[ir.Annotation]=[], engine: Optional[ir.Engine]=None):
    return ir.Update(engine, relation, tuple(args), effect, FrozenOrderedSet(annos))

def output(values: PySequence[Union[ir.Value, Tuple[str, ir.Value]]], keys: Optional[PySequence[ir.Var]] = None, engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    """Create an output task that will return values bound to these variables. The values
    sequence can contain plain Values, in which case the alias used in the output is the
    variable name. This alias can be customized by adding a tuple (alias, Value) to values.
    """
    s = ordered_set()
    for x in values:
        if isinstance(x, ir.Var):
            s.add((x.name, x))
        else:
            s.add(x)
    return ir.Output(engine, s.frozen(), keys if keys is None else FrozenOrderedSet.from_iterable(keys), FrozenOrderedSet(annos))


def construct(var: ir.Var, bindings: dict[ir.Relation, Any], engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    """Create a Construct node that will create an id and bind to this var. The id will be a
    hash of the types of the var, followed by the name and value of the bindings."""
    values = []
    values.append(var.type)
    for relation, value in bindings.items():
        values.append(ir.Literal(types.String, relation.name))
        if isinstance(value, ir.Node):
            values.append(value)
        else:
            values.append(lit(value))
    return ir.Construct(engine, tuple(values), var, FrozenOrderedSet(annos))

def aggregate(aggregation: ir.Relation, projection: PySequence[ir.Var], group: PySequence[ir.Var], args: PySequence[Any], engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Aggregate(engine, aggregation, tuple(projection), tuple(group), tuple(args), FrozenOrderedSet(annos))

def rank(projection: PySequence[ir.Var], group: PySequence[ir.Var], args: PySequence[ir.Var], arg_is_ascending: PySequence[bool], result: ir.Var, limit: int = 0, engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Rank(engine, tuple(projection), tuple(group), tuple(args), tuple(arg_is_ascending), result, limit, FrozenOrderedSet(annos))


#
# Logical Quantifiers
#

def not_(task: ir.Task, engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return Not(task, engine, annos)

def Not(task: ir.Task, engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Not(engine, task, FrozenOrderedSet(annos))

def exists(vars: PySequence[ir.Var], task: ir.Task, engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Exists(engine, tuple(vars), task, FrozenOrderedSet(annos))

def for_all(vars: PySequence[ir.Var], task: ir.Task, engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.ForAll(engine, tuple(vars), task, FrozenOrderedSet(annos))


#
# Iteration (Loops)
#

# loops body until a break condition is met
def loop(iter: ir.Var, body: ir.Task, hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Loop(engine, tuple(hoisted), iter, body, FrozenOrderedSet(annos))

def break_(check: ir.Task, engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return Break(check, engine, annos)

def Break(check: ir.Task, engine: Optional[ir.Engine]=None, annos: PySequence[ir.Annotation]=[]):
    return ir.Break(engine, check, FrozenOrderedSet(annos))



import logging
from typing import Optional

import relationalai.early_access.builder as qb
from relationalai.early_access.dsl.orm.constraints import Constraint, Unique, Mandatory, RoleValueConstraint, \
    InclusiveSubtypeConstraint, ExclusiveSubtypeConstraint
from relationalai.early_access.dsl.orm.relationships import Role, Relationship
from relationalai.early_access.metamodel.util import OrderedSet
from relationalai.early_access.dsl.core import warn


class OntologyReasoner:
    def __init__(self):

        # TODO : populate and use
        self._value_types: OrderedSet[qb.Concept] = OrderedSet()
        self._entity_types: OrderedSet[qb.Concept] = OrderedSet()
        self._exclusive_entity_types: OrderedSet[qb.Concept] = OrderedSet()

        self._concept_identifiers: dict[qb.Concept, OrderedSet[Unique]] = {}
        self._subtype_identifiers: dict[qb.Concept, OrderedSet[Unique]] = {}
        self._supertype_identifiers: dict[qb.Concept, OrderedSet[Unique]] = {}
        self._constraint_identifies: dict[Unique, qb.Concept] = {}
        self._constructor_roles: OrderedSet[Role] = OrderedSet()
        self._mandatory_roles: OrderedSet[Role] = OrderedSet()
        self._role_unique_constraints: dict[Role, OrderedSet[Unique]] = {}

        self._internal_unique_constraints: OrderedSet[Unique] = OrderedSet()
        self._role_value_constraints: OrderedSet[RoleValueConstraint] = OrderedSet()
        self._inclusive_subtype_constraints: OrderedSet[InclusiveSubtypeConstraint] = OrderedSet()
        self._exclusive_subtype_constraints: OrderedSet[ExclusiveSubtypeConstraint] = OrderedSet()

        self._subtype_map: dict[qb.Concept, OrderedSet[qb.Concept]] = {}
        self._supertype_map: dict[qb.Concept, OrderedSet[qb.Concept]] = {}
        self._subtype_closure: dict[qb.Concept, OrderedSet[qb.Concept]] = {}
        self._supertype_closure: dict[qb.Concept, OrderedSet[qb.Concept]] = {}
        self._type_closure: dict[qb.Concept, OrderedSet[qb.Concept]] = {}

        # TODO: populate to check in weaver
        self._errors: OrderedSet[Exception] = OrderedSet()

    
    #=
    # When concepts are added, update the type hierarchy and capture the preferred IDs (inferring as necessary).
    #=

    def new_concept(self, concept: qb.Concept):
        if concept._is_primitive():
            self._value_types.add(concept)
        else:
            self._entity_types.add(concept)
            self._update_type_hierarchy(concept)
            self._update_type_closures(concept)
            self._inherit_parent_identifiers(concept)

    #=
    # When constraints are added, capture the ones that are reference schemes.
    #=

    def new_constraint(self, constraint: Constraint):
        if isinstance(constraint, Unique):
            self._process_uniqueness_constraint(constraint)
        elif isinstance(constraint, Mandatory):
            self._process_mandatory_constraint(constraint)
        elif isinstance(constraint, RoleValueConstraint):
            self._role_value_constraints.add(constraint)
        elif isinstance(constraint, InclusiveSubtypeConstraint):
            self._inclusive_subtype_constraints.add(constraint)
        elif isinstance(constraint, ExclusiveSubtypeConstraint):
            self._exclusive_subtype_constraints.add(constraint)
        else:
            logging.warning(f"Unknown constraint type: {type(constraint)}")

    #=
    # Constraints: uniqueness and reference schemes.
    #=

    def ref_schemes_of(self, entity_type: qb.Concept, shallow=False) -> OrderedSet[Unique]:
        """
        Returns the reference schemes that are defined for the entity type (plus those inferred from sub- or supertypes).
        """
        own_reference_schemes = self._concept_identifiers.get(entity_type) or OrderedSet()
        if shallow or self.is_exclusive_entity_type(entity_type):
            supertype_reference_schemes = self._supertype_identifiers.get(entity_type) or OrderedSet()
            subtype_reference_schemes = self._subtype_identifiers.get(entity_type) or OrderedSet()
        else:
            supertype_reference_schemes = self.subtype_ref_schemes_of(entity_type)
            subtype_reference_schemes = OrderedSet()
        return supertype_reference_schemes | own_reference_schemes | subtype_reference_schemes

    def subtype_unique_ref_scheme_role(self, entity_type: qb.Concept) -> Role:
        """
        Returns the role used to construct the entity type.

        Only valid for simple reference schemes inferred by a subtype.
        """
        ref_schemes = self.ref_schemes_of(entity_type)
        if len(ref_schemes) != 1:
            raise ValueError(f"No unique reference scheme found for {entity_type}")
        ref_scheme = ref_schemes[0]
        return self._extract_unique_role_from_uc(ref_scheme)

    def subtype_ref_schemes_of(self, entity_type: qb.Concept) -> OrderedSet[Unique]:
        """
        Returns the reference schemes that are defined for the entity type and any of its parent types.
        """
        ref_schemes = entity_type._ref_scheme()
        if ref_schemes is None:
            raise ValueError(f"No reference scheme found for {entity_type}")
        identifier_constraints = OrderedSet()
        for relationship in ref_schemes:
            if not isinstance(relationship, Relationship):
                raise ValueError(f"Expected a Relationship, but got {type(relationship)}")
            identifier_constraints.add(self.lookup_identifier_constraint(relationship))
        return identifier_constraints

    def own_ref_scheme_of(self, entity_type: qb.Concept) -> Unique:
        """
        Returns the reference schemes that are defined for the entity type (but not inferred).
        """
        own_ref_schemes = self._concept_identifiers[entity_type]
        if len(own_ref_schemes) != 1:
            raise ValueError(f"Expected exactly one reference scheme for {entity_type}, but got {len(own_ref_schemes)}")
        return own_ref_schemes[0]

    def own_ref_scheme_role(self, entity_type: qb.Concept) -> Role:
        """
        Returns the role used to construct the entity type. Only valid for simple reference schemes.
        """
        ref_scheme = self.own_ref_scheme_of(entity_type)
        return self._extract_unique_role_from_uc(ref_scheme)

    def has_own_ref_scheme(self, entity_type: qb.Concept) -> bool:
        """
        Returns true if the entity type has a reference scheme defined for it.
        """
        return entity_type in self._concept_identifiers

    def has_simple_ref_scheme(self, entity_type: qb.Concept) -> bool:
        ref_schemes = self.ref_schemes_of(entity_type)
        if len(ref_schemes) == 0:
            raise ValueError(f"No reference scheme found for {entity_type}")
        for uc in ref_schemes:
            if len(uc.roles()) > 1:
                return False
        return True

    def has_composite_ref_scheme(self, entity_type: qb.Concept) -> bool:
        return not self.has_simple_ref_scheme(entity_type)

    def is_constructing_role(self, role: Role) -> bool:
        """
        Returns true if the role is a constructing role (i.e., it is used to construct an entity type).
        """
        return role in self._constructor_roles

    def concept_identifiers(self) -> dict[qb.Concept, OrderedSet[Unique]]:
        """
        Returns a mapping of entity types to their reference schemes.
        """
        return self._concept_identifiers

    def is_identifier_relationship(self, relationship: Relationship) -> bool:
        """
        Returns true if the relationship is an identifier relationship (i.e., it is used to construct an entity type).
        """
        if not relationship._binary():
            return False
        for role in relationship._roles():
            if self.is_constructing_role(role):
                return True
        return False

    def lookup_identifier_constraint(self, relationship: Relationship) -> Unique:
        """
        Returns the corresponding identifier constraint for an identifier relationship.
        """
        if not self.is_identifier_relationship(relationship):
            raise ValueError(f"Expected an identifier relationship, but got {relationship}")
        concept_role = None
        constructing_role = None
        for role in relationship._roles():
            if self.is_constructing_role(role):
                concept_role = role.sibling()
                constructing_role = role
            else:
                constructing_role = role.sibling()
                concept_role = role
        assert concept_role is not None and constructing_role is not None, \
            f"Expected both roles to be defined in {relationship}"
        for constraint in self._concept_identifiers[concept_role.player()]:
            if constructing_role in constraint.roles():
                return constraint
        raise ValueError(f"No identifier constraint found for {relationship}")

    def identifies_concept(self, constraint: Unique) -> qb.Concept:
        """
        Returns the concept that is identified by the given uniqueness constraint.
        """
        if constraint not in self._constraint_identifies:
            raise KeyError(f"No concept identified by {constraint}")
        return self._constraint_identifies[constraint]

    def is_composite_concept(self, concept: qb.Concept) -> bool:
        """
        Returns true if the concept is a composite concept (i.e., it has a composite reference scheme).
        """
        ref_schemes = self.ref_schemes_of(concept)
        if len(ref_schemes) == 0:
            return False
        for uc in ref_schemes:
            if len(uc.roles()) > 1:
                return True
        return False

    def is_exclusive_entity_type(self, entity_type: qb.Concept) -> bool:
        """
        Returns true if the entity type is an exclusive entity type).
        """
        return entity_type in self._exclusive_entity_types or self._check_exclusive_supertype(entity_type)

    def _check_exclusive_supertype(self, entity_type: qb.Concept) -> bool:
        """
        Checks if the entity type is an exclusive supertype.
        """
        if self.has_own_ref_scheme(entity_type):
            return False
        subtypes = self._subtype_map.get(entity_type)
        is_exclusive = subtypes is not None
        if is_exclusive:
            self._exclusive_entity_types.add(entity_type)
        return is_exclusive

    def subtype_exclusive_supertype(self, subtype: qb.Concept) -> Optional[qb.Concept]:
        """
        Returns the exclusive supertype of the given subtype if it exists.
        """
        supertypes = self._supertype_map.get(subtype)
        if not supertypes:
            return None
        for supertype in supertypes:
            if self.is_exclusive_entity_type(supertype):
                return supertype
        return None

    def is_mandatory_role(self, role: Role) -> bool:
        """
        Returns true if the role is a mandatory role.
        """
        return role in self._mandatory_roles

    def is_one_role(self, role: Role) -> bool:
        """
        Returns true if the role is a "one" role (as opposed to a "many" role),
          i.e. it is not spanned by an internal uniqueness constraint.
        """
        return not any(c in self._internal_unique_constraints for c in self._role_unique_constraints.get(role, OrderedSet()))

    # [REKS: TODO] Need to loosen this requirement so that we admit
    #              compatible roles when the player of r1 is a subtype
    #              of that of r2 or vice versa.
    def are_roles_compatible(self, role1: Role, role2: Role) -> bool:
        warn(f"Checking compatibility of roles played by {str(role1.player())} and {str(role2.player())} without reasoning about subtypes")
        return role1.player() == role2.player()

    # [REKS: TODO] Need to do the subtype/supertype reasoning required
    #              to properly answer this w/o requring c1 to be equal
    #              to c2
    def least_supertype(self, concept1: qb.Concept, concept2: qb.Concept) -> Optional[qb.Concept]:
        warn(f"Trying to find least supertype of {str(concept1)} and {str(concept2)} without reasoning about subtypes")
        return None if str(concept1) != str(concept2) else concept1

    @staticmethod
    def _extract_unique_role_from_uc(uc: Unique) -> Role:
        """
        Extracts the unique role from a uniqueness constraint.
        """
        roles = uc.roles()
        if len(roles) != 1:
            raise ValueError(f"Expected exactly one role in uniqueness constraint, but got {len(roles)}")
        return roles[0]


    def _update_type_hierarchy(self, new_entity_type: qb.Concept):
        parent_concepts = new_entity_type._extends
        for parent_concept in parent_concepts:
            self._subtype_map.setdefault(parent_concept, OrderedSet()).add(new_entity_type)
            self._supertype_map.setdefault(new_entity_type, OrderedSet()).add(parent_concept)

    def _update_type_closures(self, new_entity_type: qb.Concept):
        parent_concepts = new_entity_type._extends
        self._supertype_closure.setdefault(new_entity_type, OrderedSet()).update(parent_concepts)
        for parent_concept in parent_concepts:
            self._supertype_closure[new_entity_type].update(self._supertype_closure[parent_concept])
        self._type_closure.setdefault(new_entity_type, OrderedSet()).update(self._supertype_closure[new_entity_type])

        for supertype in self._supertype_closure[new_entity_type]:
            self._subtype_closure.setdefault(supertype, OrderedSet()).add(new_entity_type)
            self._type_closure.setdefault(supertype, OrderedSet()).add(new_entity_type)

    def _inherit_parent_identifiers(self, new_entity_type: qb.Concept):
        for parent in new_entity_type._extends:
            self._supertype_identifiers.setdefault(new_entity_type, OrderedSet()).update(
                self._supertype_identifiers.get(parent, OrderedSet())).update(
                    self._concept_identifiers.get(parent, OrderedSet()))

            
    def _process_uniqueness_constraint(self, constraint: Unique):
        for role in constraint.roles():
            self._role_unique_constraints.setdefault(role, OrderedSet()).add(constraint)

        if len(set(r._part_of() for r in constraint.roles())) == 1:
            self._internal_unique_constraints.add(constraint)
        
        # note: ignoring the ones that are not preferred identifiers for now
        if not constraint.is_preferred_identifier:
            return

        roles = constraint.roles()
        self._constructor_roles.update(roles)  # mark as ctor roles

        constructed_concept = None
        if constraint._is_internal():
            # simple ref scheme
            constructed_concept = self._process_simple_ref_scheme(constraint)
        else:
            # composite ref scheme
            constructed_concept = self._process_composite_ref_scheme(constraint)
        self._update_subtype_identifiers(constructed_concept, constraint)

    def _process_simple_ref_scheme(self, constraint: Unique):
        role = constraint.roles()[0]
        relation = role._relationship  # actually a DSL Relation
        if relation._arity() != 2:
            raise ValueError(f"Identifier relationship {relation} should have arity 2, but got {relation._arity()}")
        sibling_role = role.sibling()
        if sibling_role is None:
            raise ValueError(f"Unable to find the sibling role for {role}")
        constructed_concept = sibling_role.player()
        self._concept_identifiers.setdefault(constructed_concept, OrderedSet()).add(constraint)
        self._constraint_identifies[constraint] = constructed_concept
        return constructed_concept

    def _process_composite_ref_scheme(self, constraint: Unique):
        roles = constraint.roles()
        if len(roles) <= 1: 
            raise ValueError(f"External uniqueness constraint should have more than one role, but got {len(roles)}")
        role = roles[0]  # take an arbitrary role to get the constructed concept
        sibling_role = role.sibling()
        if sibling_role is None:
            raise ValueError(f"Unable to find the sibling role for {role}")
        constructed_concept = sibling_role.player()
        self._concept_identifiers.setdefault(constructed_concept, OrderedSet()).add(constraint)
        self._constraint_identifies[constraint] = constructed_concept
        return constructed_concept
    
    def _update_subtype_identifiers(self, entity_type: qb.Concept, constraint: Unique):
        # Clear subtype identifiers for the entity type that now has its own, if they exist
        self._subtype_identifiers.pop(entity_type, None)

        # For all subtypes that now have their own (or parent's) identifiers, clear subtype identifiers
        subtypes = self._subtype_closure.get(entity_type, OrderedSet())
        for subtype in subtypes:
            if subtype._ref_scheme():
                self._subtype_identifiers.pop(subtype, None)

            self._supertype_identifiers.setdefault(subtype, OrderedSet()).add(constraint)

        # For all supertypes that do not have their own (or parent's) identifiers, add the current constraint
        supertypes = self._supertype_closure.get(entity_type, OrderedSet())
        for supertype in supertypes:
            if not supertype._ref_scheme():
                self._subtype_identifiers.setdefault(supertype, OrderedSet()).add(constraint)

    def _process_mandatory_constraint(self, constraint: Mandatory):
        role = constraint.roles()[0]
        self._mandatory_roles.add(role)



from relationalai.early_access.builder import Integer
from relationalai.early_access.builder import define, max

from .utilities import clone_relation

def linear_prefix_sum(list):
    result = list._model.Relationship("result1 {Integer} {Integer}")

    i, val, val1, val2 = Integer.ref(), Integer.ref(), Integer.ref(), Integer.ref()

    define(result(i, val)).where(
        list(i, val),
        i == 1
    )

    define(result(i, val)).where(
        result(i - 1, val1),
        list(i, val2),
        val == val1 + val2
    )

    return result

def linear_prefix_sum_with_groupby(r):
    result = clone_relation(r)
    i, n, val, val1, val2 = Integer.ref(), Integer.ref(), Integer.ref(), Integer.ref(), Integer.ref()

    if r._arity() == 3:
        u = r._field_refs[0]  # Assuming the first field is the grouping field

        define(result(u, 1, val)).where(
            r(u, 1, val)
        )

        define(result(u, i, val)).where(
            result(u, i - 1, val1),
            r(u, i, val2),
            val == val1 + val2
        )
    else:
        # Assuming the first two fields are the grouping fields
        u = r._field_refs[0]
        n = r._field_refs[1]

        define(result(u, n, 1, val)).where(
            r(u, n, 1, val)
        )

        define(result(u, n, i, val)).where(
            result(u, n, i - 1, val1),
            r(u, n, i, val2),
            val == val1 + val2
        )
        
    return result


def dynamic_prefix_sum(list):
    result = list._model.Relationship("result2 {Integer} {Integer}")
    L = list._model.Relationship("L1 {Integer} {Integer} {Integer}")
    B = list._model.Relationship("B1 {Integer} {Integer} {Integer}")
    B_max = list._model.Relationship("B_max1 {Integer} {Integer}")

    i, j, k = Integer.ref(), Integer.ref(), Integer.ref()
    val, val1, val2 = Integer.ref(), Integer.ref(), Integer.ref()

    define(B(i, 1, val)).where(list(i, val))
    define(B(i, j, val)).where(
        B(i, k, val1),
        B(i - k, k, val2),
        val == val1 + val2,
        j == 2 * k,
        i % j == 0
    )

    define(L(i, j, val)).where(
        B_max(i, j),
        B(i, j, val),
    )

    define(B_max(i, max(j).per(i))).where(
        B(i, j, k)
    )

    define(result(i, val)).where(L(i, i, val))
    define(result(i, val)).where(
        L(i, j, val1),
        result(i - j, val2),
        val == val1 + val2
    )

    return result


from typing import Any

from relationalai.early_access.builder import annotations, Relationship as QBRelationship


class AssertedRelation(QBRelationship):
    """
    A relation that's assumed to be defined in the target system, but cannot be verified.

    This is used as a common parent class for such relations, and the compiler makes sure to
    not generate any unnecessary constructs for them.
    """

    def __init__(self, model, name: str, *sig: Any):
        args = [f'{{_x_{idx}:{typ}}}' for idx, typ in enumerate(sig)]
        madlib = f'{name} {" ".join(args)}'
        super().__init__(madlib, short_name=name, model=model.qb_model())
        self.annotate(annotations.external)
        self._dsl_model = model



"""
Product graph implementation for automaton-based path finding.
"""

from relationalai.early_access.builder import Integer, where, define, Relationship, String

from .graph import Graph


def create_product_graph(graph, automaton):
    """
    Create a product graph based on an automaton specification.

    Args:
        model: The RelationalAI model
        automaton_spec: List of automaton transitions (src_state, dst_state, label/filter)
        node: Node concept from the original graph
        edge: Edge relationship from the original graph

    Returns:
        Tuple of (ProductGraphNode, product_graph) representing the product graph concepts
    """
    model = graph.model
    Node = graph.Node
    Edge = graph.Edge

    # Define product graph for path finding
    ProductGraphNode = model.Concept("ProductNode")
    ProductGraphEdge = model.Relationship(f"pg_edge {{{ProductGraphNode}}} {{{ProductGraphNode}}}")

    ProductGraphNode.orig_node = Relationship(f"orig_node {{{ProductGraphNode}}} {{{Node}}}")
    ProductGraphNode.state = Relationship(f"state {{{ProductGraphNode}}} {{Integer}}")
    ProductGraphNode.identify_by(ProductGraphNode.state, ProductGraphNode.orig_node)

    # NB: Modifying `automaton` object
    # Add automaton transitions to the model
    automaton.transition = model.Relationship("automaton {Integer} {Integer} {String}")

    # Create references for use in rules
    state1, state2 = Integer.ref(), Integer.ref()
    node1, node2 = Node.ref(), Node.ref()

    # Create a rule for each automaton transition:
    for state1 in automaton.delta:  # delta is the dict containing automaton transitions
        for state2 in automaton.delta[state1]:
            for transition in automaton.delta[state1][state2]:
                label = transition.get_label().label

                define(automaton.transition(state1, state2, label))

                source_filters = transition.src_filters

                where(
                    ProductGraphNode.state == state1,
                    ProductGraphNode.orig_node == node1,
                    Edge(node1, node2, label),
                    *([f(node1) for f in source_filters])
                    ).define(
                        ProductGraphNode.new(state=state2, orig_node=node2),
                        ProductGraphEdge(ProductGraphNode, ProductGraphNode.new(state=state2, orig_node=node2))
                    )

    return Graph(model, ProductGraphNode, ProductGraphEdge)


def project_product_graph_paths_to_original_graph(product_graph, product_graph_paths, automaton):

    model = product_graph.model
    ProductGraphNode = product_graph.Node

    # extract Node type of original graph:
    Node = ProductGraphNode.orig_node._fields[1].type_str

    # print("Node type in original graph:", Node)


    # Project path down from product graph to original graph
    projected_paths = model.Relationship(f"projected {{Integer}} {{Integer}} {{String}} {{{Node}}}")

    k = Integer.ref()
    product_graph_node2 = ProductGraphNode.ref()
    label = String.ref()

    i = Integer.ref()

    define(projected_paths(i, 0, "", ProductGraphNode.orig_node)).where(product_graph_paths(i, 0, ProductGraphNode))
    define(projected_paths(i, k, label, product_graph_node2.orig_node)).where(
        product_graph_paths(i, k - 1, ProductGraphNode),
        product_graph_paths(i, k, product_graph_node2),
        automaton.transition(ProductGraphNode.state, product_graph_node2.state, label)
    )

    return projected_paths